/*
 * $Id: corpserv.c,v 1.10 1996/05/17 14:32:49 scotty Rel $
 *
 * $Log: corpserv.c,v $
 * for world edition: last change was to increase bigbuff 4000 to 6000
 * Revision 1.10  1996/05/17 14:32:49  scotty
 * Caught the last remnant of the old logging scheme which was
 * trying to fflush() an unitialized FILE pointer.
 *
 * Revision 1.9  1996/05/13 01:23:55  scotty
 * v927 fixes from Tony.
 *
 * Revision 1.8  1996/05/12 22:13:55  scotty
 * Added a signal handler to catch children when they exit;
 * modified the call to accept() to handle EINTRs caused by
 * the handler.
 *
 * Revision 1.7  1996/03/01 23:54:11  scotty
 * Fixed call to setsockopt() -- the wrong size was being passed and
 * some compilers want a cast for &one.
 *
 * __isascii is now only defined if it's not already defined
 * (Linux defines it).
 *
 * Revision 1.6  1996/02/26 22:51:30  scotty
 * Added support for braindead (Ultrix) syslog()s.
 *
 * Revision 1.5  1996/02/26 21:05:57  scotty
 * Moved idxpath[] into here from corphash.c since it's only every
 * used in this file.
 *
 * Revision 1.4  1996/02/26 19:35:25  scotty
 * Most exit()s now print a message explaing why the server has quit.
 *
 * Revision 1.3  1996/02/15 01:26:38  scotty
 * Now writes the server PID in ACCPATH/sarad.pid.
 *
 * Revision 1.2  1996/02/14 22:01:04  scotty
 * Some fixes by Tony.
 *
 * Revision 1.1  1996/01/30 20:51:18  scotty
 * Initial revision
 *
 */
#include "corpinc.h"

static char rcsid[] = "$Id: corpserv.c,v 1.10 1996/05/17 14:32:49 scotty Rel $";

#undef isspace

#define SERVER 1
#define MAX_USER 20
#define MAX_PW 20
#define MAX_NAME 8
#define MAX_CQL 2000
#define MAX_REGEXP 20
#define MAX_FILTER 20
#define MAX_SOL (MAXSENTENCE+40)
#define MAXFILTERN 100

struct ran {
   int       min;
   int       size;
};


union uarg {
   char        user[MAX_USER];
   char        pw[MAX_PW];
   char        name[MAX_NAME];
   int         num;
   sara_long   lnum;
   char        cql[MAX_CQL];
   char        regexp[MAX_REGEXP];
   int         (*table)[];
   char        filter[MAX_FILTER];
   struct ran  range;
};


#define MAX_COMMAND 200
#define MAX_OPS 5

struct pcmd {
   int       i;
   union uarg arg[MAX_OPS];
};


struct cmd {
   char     *name;
   int       op[MAX_OPS];
};

#define NFILES 40

struct qt {
   BOOL      bUse;
   BOOL      bSave;
   char      name[12];
   char      path[_MAX_PATH];
} qtable[NFILES];


#define OP_USER    1
#define OP_PW      2
#define OP_NAME    3
#define OP_NUM     4
#define OP_CQL     5
#define OP_REGEXP  6
#define OP_TABLE   7
#define OP_FILTER  8
#define OP_RANGE   9
#define OP_LNUM   10


struct cmd ctable[] = {
   { "LOGOUT",       { 0,        0,          0,          0        }},
   { "LOG",          { OP_USER,  OP_PW,      0,          0        }},
   { "SAVE",         { OP_NUM,   OP_NAME,    0,          0        }},
   { "DIR",          { OP_NAME,  0,          0,          0        }},
   { "REMOVE",       { OP_NAME,  0,          0,          0        }},
   { "PURGE",        { OP_NUM,   0,          0,          0        }},
   { "SOLVE",        { OP_NAME,  OP_CQL,     0,          0        }},
   { "SUBCORPUS",    { OP_NAME,  OP_CQL,     0,          0        }},
   { "WORDLIST",     { OP_NAME,  OP_REGEXP,  0,          0        }},
   { "SQTABLE",      { OP_NAME,  OP_NAME,    OP_TABLE,   0,       }},
   { "FILTER",       { OP_NAME,  OP_FILTER,  0,          0        }},
   { "SORTFILTER",   { OP_NAME,  OP_NUM,     OP_FILTER,  0        }},
   { "SORT",         { OP_NAME,  OP_NUM,     0,          0        }},
   { "GET",          { OP_NAME,  OP_NUM,     OP_CQL,     0        }},
   { "CUT",          { 0,        0,          0,          0        }},
   { "CHAR",         { OP_NUM,   OP_NAME,    0,          0        }},
   { "GETPOS",       { OP_NAME,  0,          0,          0        }},
   { "GETDTD",       { OP_NUM,   0,          0,          0        }},
   { "MAXLENGTH",    { OP_NUM,   0,          0,          0        }},
   { "GETSC",        { OP_NAME,  OP_NUM,     0,          0        }},
   { "GETHEAD",      { OP_NAME,  OP_LNUM,    OP_NUM,     0        }},
   { "LOOKUP",       { OP_CQL,  0,          0,          0        }},
   { "DMATCH",       { OP_NUM,   0,          0,          0        }},
   { "QNAME",        { 0,        0,          0,          0        }},
   { "OPEN",         { OP_NAME,  0,          0,          0        }},
   { "CSCORE",       { OP_NAME,  OP_NUM,     OP_CQL,     0        }},
   { "LOC",          { OP_NAME,  OP_NUM,     0,          0        }},
   { "GETHEAD2",     { OP_NAME,  OP_LNUM,    OP_LNUM,    OP_NUM   }},
   { "GETCHEAD",     { OP_NUM,   0,          0,          0        }},
   { "GET1SOL",      { OP_NAME,  OP_NAME,    OP_NAME,    0        }},
   { "MOTD",         { 0,        0,          0,          0        }},
   { "PWD",          { OP_PW,    OP_PW,      0,          0        }},
   { "INFO",         { OP_NAME,  0,          0,          0        }},
   { "RLOOKUP",      { OP_CQL,   0,          0,          0        }},
   { "RGET",         { OP_NUM,   0,          0,          0        }},
   { "BIB",          { OP_NAME,  0,          0,          0        }},
   { "BIBITEM",      { OP_NAME,  OP_NUM,     0,          0        }},
   { "TRACE",        { OP_NUM,   0,          0,          0        }},
   { "DOWNLOAD",     { OP_NAME,  0,          0,          0        }},
   { "LINE",         { 0,        0,          0,          0        }},
   { "THIN",         { OP_NAME,  OP_NAME,    OP_NUM,     OP_NUM   ,OP_NUM}},
   { "WEB" ,         { 0,        0,          0,          0,       0}},
   { "SCDEF",			{OP_NAME,0,			0,			0,		0}},
   { "SCADD",			{OP_NAME, OP_TABLE,0,0,0}},
   { "SCSET",			{OP_NAME}},
   {"FTAB", {OP_NUM,OP_NUM,OP_NUM,OP_CQL}},
   {"FENTRY", {OP_NUM}},
   {"FFREE",{0}},
   {"SCSAVE",{0}},
   {"SCLIST",{OP_NUM}},
   {"SCENUM",{OP_NAME, OP_NUM, OP_NUM}},
   {"CTAB",{OP_NAME,OP_NUM,OP_NUM,OP_CQL}},
   {"CTABENTRY", {OP_NUM}},
   {"ACSCORE", {OP_NAME,OP_NUM,OP_NUM,OP_NUM,OP_CQL}},
   {"CTABTHIN",{OP_NAME,OP_NAME,OP_NUM,OP_NUM,OP_CQL}},
   {"CTABOPTIONS",{OP_NUM,OP_NUM,OP_NAME,OP_NUM}},
   {"FORM",{OP_NUM,OP_NAME}},
   {"LEMMINFO", {0}},
   {"LEMMALL", {0}},
   {"LEMMSEL", {OP_NAME}},
   { "GETSOL",          { OP_NAME,  OP_NUM,     OP_CQL,     0        }},
   {"CTABFREE",{0}},
   {"LEMMDESC",{OP_NAME}},
   {"LEMMTEST",{OP_NAME, OP_NAME, OP_NAME, OP_NUM}},
};


struct cmd ftable[] = {
   { "NOSGMLX",   { 0,        0, 0, 0  }},
   { "NOPOSX",    { 0,        0, 0, 0  }},
   { "NORMSPACE", { 0,        0, 0, 0  }},
/*   { "LEFTL",     { OP_NUM,   0, 0, 0  }},
   { "ADDSLEN",   { OP_NUM,   0, 0, 0  }}, */
   { "CMAP",      { 0,        0, 0, 0  }},
/*   { "ADDPOS",    { OP_NUM,   0, 0, 0  }},
   { "ADDTEXT",   { OP_NUM,   0, 0, 0  }}, */
   { "NORMCR0",   { 0,        0, 0, 0  }},
   { "ADJSGML",   { 0,        0, 0, 0  }},
   { "ADJPOS",    { 0,        0, 0, 0  }},
   { "WTOPOS",    { 0,        0, 0, 0  }},
};


enum filtype {
   FIL_SGMLX, FIL_NOPOSX, FIL_NORMSPACE,
   /* FIL_LEFTL, FIL_ADDSLEN, */ FIL_CMAP, /* FIL_ADDPOS,
   FIL_ADDTEXT, */FIL_NORMCR0, FIL_ADJSGML, FIL_ADJPOS,
   FIL_WTOPOS
};


char        idxpath[_MAX_PATH];

int         nPOSx;

BOOL        bSLen = FALSE;
BOOL        bPOS = FALSE;
BOOL        bText = FALSE;

BOOL bNoBib;

int         iq = 0;

FILE       *fSysLog;  
#ifdef __unix__
FILE *fAdmLog;
#endif

char        calluser[MAX_USER];
char        callloc[30] = "local";

BOOL        bInt;

BOOL bSkipTests=FALSE;
/*
 * LOGOUT
 * LOG user password
 * SAVE name
 * DIR
 * DIR q
 * REMOVE q
 * PURGE n
 * SOLVE name query
 * SUBCORPUS name query
 * WORDLIST name pattern
 * SQTABLE name name data
 * FILTER name filter
 * SORTFILTER name n filter
 * SORT name n
 * GET name n,m
 * CUT
 */


#define NCOMMAND (sizeof(ctable)/sizeof(struct cmd))

char        psent[MAX_SOL];

void        GetCString(char *);
BOOL        ParseString(char *, struct pcmd *, struct cmd[]);
void        ExecString(struct pcmd *);
void        SetString(char *);
BOOL        LookUpCmd(char *, int *, struct cmd[]);
void        GetArg(int, char **, union uarg *);
void        GetBTS(char **, char *);
void        GetRest(char **, char *,int);
void        GetTable(char **, int *);
void        GetRange(char **, struct ran *);
void        parg(int, union uarg *);
void        Log(char *, char *);
void        Logout(void);
BOOL        NewQName(char *);
char       *GetQPath(char *);
void        FreeQPath(char *);
char       *GetRef(sara_ulong, sara_long, WORDNO, int,int, int *, int *,
                   char *scope, char *);
char       *GetSol(char *, int, int *, int *, sara_long *, char *, char *);
void        BoundElement(char *, WORDNO, sara_ulong, sara_long *,
                         sara_long *, int *, sara_long *, sara_long,
                         sara_long);
void        BoundElement2(char *, WORDNO, sara_ulong, sara_long *,
                          sara_long *, int *, sara_long *, sara_long,
                          sara_long);
void        AddFilter(char *, struct pcmd *);
void        ClearFilters(char *);
void        Reformat(char *, char *, int *, int *);
/* void        ApplyFilter(struct pcmd *, char *, int *, int *); */
void        ApplyFilter(enum filtype, char *, int *, int *);
void        StripAllSGML(char *s, int *, int *);
void        DelSGML(char *, int *, int *);
int         IsSGML(char *);
void        StripAllPOS(char *s, int *, int *);
int         DelPOS(char *, int *, int *);
int         IsPOS(char *, int *);
BOOL        strtest(char *, char *);
void        NormSpace(char *, int *, int *);
void        DelNL(char *, int *);
void        ConsolSpaces(char *, int *);
void        LeftL(char *, int, int *, int *);
void        AddChar(char *, int);
void        ClearChars(void);
void        TranslateChars(char *, int *, int *);
int         IsChar(char *, int *);
void        SQTable(char *,char *, int, int *);
void        Thin(char *, char *, int, int, int, int *, int *);
void        ThinInit(char *, char *, int, int *);
void        ThinRand(char *, char *, int, int *);
void        OnePerText(char *, char *, int *);
void        NormCR0(char *, int *, int *);
char       *GetHeadElt(char *, sara_long, int, sara_long *, int *,
                       BOOL *, sara_long *);
int         GetElt(FILE *, char *, int *, char *);
BOOL        Empty(char *);
sara_long   FindHeader(FILE *);
BOOL        Ignore(char *);
BOOL        AllBlank(char *);
void        ETag(FILE *, sara_long *);
void        Trim(char *, int *, int *);
BOOL Startup(char*);
#if defined(SERVER)
int         readrec(FILE *, char *, char *, char *);
BOOL        readst(FILE *, char *);
#endif /* ! SERVER */

BOOL        GetLims(FILE *, sara_long, sara_long, const char *,
                    sara_long *, sara_long *);
void        GetDictEntry(FILE *, sara_long, char *, int, sara_long *, sara_long*);
void LowerCase(char*);
BOOL NewSubCorpus(const char*);
BOOL SetSubCorpus(const char*);
BOOL AddSubCorpus(const char*,int, int*);
BOOL SCSave(char*);
char* SCList(int);
char* SCEnum(const char*,int,int);
int SCCount();
void SCReadAll();
BOOL SCRegistered();
int FTable(int,int,int,const char*);
void FEntry(int,char*,int*,int*);
void FFree();
int QuickFrequency(const char*);
int DicFrequency(const char*);
void GetFTable(char*);
void CTab(int,char**,int*, double*);
int MakeCTab(char*,int,int,char*);
void ACScore(char*,int,int, int,char*,long*,double*);
void CTabOptions(int,int,char*,int);
BOOL Form(char*,int,char*,int*);
char* LemmInfo();

#if defined(__unix__)
int         dcmp(signed char *, signed char *);
#else /* ! __unix__ */
int         dcmp(unsigned char *, unsigned char *);
#endif /* __unix__ */

int         OpenQPath(char *);
int         InUse(char *);
BOOL        Exists(char *);
void        Dummy(char *name);
void        GetCName(char *, char *);
void        CollTable(char *, int);
void        WToPos(char *, int *, int *);
int         IsSpec(char *, char **);
void        TranslateSpec(int, char **);
BOOL        wtag(char *);
BOOL        DTDelt(int, char *);
void CleanUp();
BOOL GetAttrib(sara_ulong,sara_ulong,char*,char*,char*);
void CTabThin(char*,char*,int,int,char*,int*,int*);
void CTabFree();

#if !defined(__unix__)
int         getpid(void);
#else
void        ALogEvent(char *,...);
#endif /* __unix__ */

BOOL        GetBibHeader(char *, int *, int *);
BOOL        GetBibItem(char *, int, char *);

#if !defined(__unix__)
# if defined(WIN32)
BOOL(*interrupt_function) (void);
# else /* ! WIN32 */
BOOL(__far __pascal * interrupt_function) (void);
# endif /* WIN32 */
#endif /* ! __unix__ */

void        LoadChars(char *);

#if defined(__unix__)
void        detach(void);
#endif /* __unix__ */

int         nLog;
int         ntry;

#define MAXTRY 3

BOOL        bServer = FALSE;
int         max_sent = MAXSENTENCE;
sara_long   dbase;

FILE       *fPack;
FILE       *fBib;
FILE       *fBibIndex;
FILE       *fDownload;

BOOL        bDTD;
int         nDTD;

struct pihead idxhead;
char      param_file[_MAX_PATH];           
BOOL bInit;

#if defined(__unix__)

void      Test();
void      StartServer();
void      XServer(char *, char *);
void      Session(int);
void      sig_alarm(int);
void      sig_urgent(int);


main(int argc, char **argv)
{
   char          *param = PARAM;
   int            opt;
   int            errflag = 0;
   int            debug = 0;
   extern int     optind;
   extern char   *optarg;
   char          *progname;
   
   optarg = NULL;

   if (progname = strrchr(*argv, '/'))
      progname++;
   else
      progname = *argv;
   
   while (!errflag && (opt = getopt(argc, argv, "dnp:")) != -1)
      switch (opt) {
       case 'd':  debug = 1;        break;
       case 'p':  param = optarg;   break;
		case 'n': bSkipTests=TRUE; break;
       default:   errflag++;        break;
      }
   
   if (errflag) {
      fprintf(stderr, "Usage: %s [-d] [-n] [-p config_file]\n", progname);
      exit(1);
   }
   
   if (!Startup(param)) {
   	printf("There were errors detected in startup tests. See testlog.txt.\n");
   	exit(1);
   }	
#if defined(LOG_MAIL)
   openlog(progname, LOG_PID|LOG_NDELAY, LOG_FACILITY);
#else
   openlog(progname, LOG_PID,LOG_FACILITY);
#endif
   
   if (debug)
      Test();
   else 
      StartServer();
}


void Test()
{
   char      s[1000];
   char      ss[1000];
   char      ap[_MAX_PATH];
   int       n;

   printf("Type end to leave test mode\n");
   syslog(LOG_NOTICE, "Started server in test mode");

   sprintf(ap, "%sadmlog", accpath);
   fAdmLog = fopen(ap, "a");
   ALogEvent("Started server in test mode");


   n = AllocBuffers(nbuff);

   syslog(LOG_INFO, "Allocated %d+%d buffers", n - nbuff, nbuff);

   while (1) {
      printf("> ");
      scanf("%[^\n]", s);
      if (!strcmp(s, "end"))
         break;

      getchar();
      XServer(s, ss);
      printf("%s\n", ss);
   }
}

struct sockaddr_in   passive,
                     inaddr;


void tryname(struct in_addr, char *);

int child;

void backup(char *path, int i)
{
   char      p1[_MAX_PATH],
             p2[_MAX_PATH];

   if (i == 0)
      strcpy(p1, path);
   else
      sprintf(p1, "%s.%d", path, i);
   sprintf(p2, "%s.%d", path, i + 1);
   remove(p2);
   rename(p1, p2);
}

void catch_child(int sig)
{
   int   status;
   pid_t pid;
   
   if (sig == SIGCHLD) {
      while (   (pid = waitpid(-1, &status, WNOHANG)) == -1
             && (errno == EINTR))
         ;
      if (WIFSIGNALED(status))
         syslog(LOG_DEBUG, "Child pid=%d was killed with signal %d",
                pid, WTERMSIG(status));
      else if (WIFEXITED(status))
         syslog(LOG_DEBUG, "Child pid=%d exited with status %d",
                pid, WEXITSTATUS(status));
   }
}

void StartServer()
{
   int               s;
   int               m;
   int               inal;
   char              us[50];
   char              us1[50];
   int               one = 1;
   char              ap[_MAX_PATH];
   int               i;
   struct servent   *sp;
   uid_t             uid;
   struct passwd    *pw;
   FILE             *fpPID;
   struct sigaction  saChild;

   saChild.sa_handler = catch_child;
   sigemptyset(&saChild.sa_mask);
   saChild.sa_flags = SA_NOCLDSTOP;

   sigaction(SIGCHLD, &saChild, 0);
                             
   GetPIndex();
   GetDict();

   sprintf(ap, "%ssyslog", accpath);
   for (i = nback - 1; i >= 0; i--)
      backup(ap, i);
   sprintf(ap, "%sadmlog", accpath);
   fAdmLog = fopen(ap, "a");

   uid = getuid();
   pw = getpwuid(uid);

   syslog(LOG_NOTICE, "Started server (%s)", pw->pw_name);
   fprintf(stdout, "Started server\n");
   fflush(stdout);

   detach();
   sprintf(ap, "%ssarad.pid", accpath);
   if ((fpPID = fopen(ap, "w")) == NULL) {
      syslog(LOG_WARNING, "Couldn't create ap", ap);
   } else {
      fprintf(fpPID, "%d\n", getpid());
      fclose(fpPID);
   }

   s = socket(AF_INET, SOCK_STREAM, 0);

   if (s < 0) {
      printf("Error opening socket\n");
      exit(1);
   }

   passive.sin_family = AF_INET;
   passive.sin_addr.s_addr = INADDR_ANY;
   passive.sin_port = htons(port);
/*
   passive.sin_port=sp->s_port;
*/

   if (bind(s, (struct sockaddr *)&passive,
            sizeof(struct sockaddr_in)) < 0) {
      fprintf(stderr, "Error binding socket [%d]", errno);
      perror("bind()");
      syslog(LOG_ERR, "Error binding socket [%d]", errno);
      exit(1);
   }

   setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *)&one, sizeof(one));
   listen(s, 1);
   while (1) {
      inal = sizeof(inaddr);

      /*
       * Server used to fall over if accept() failed
       *
       * Since it now catches children, we just log errors
       * and keep going (ignoring EINTR errors)
       */
      while ((m = accept(s, (struct sockaddr *)&inaddr, &inal)) == -1) {
         if (errno != EINTR)  {
            syslog(LOG_ERR, "Error accepting connection [%d]", errno);
         }
      }

      tryname(inaddr.sin_addr, callloc);

      syslog(LOG_INFO, "Connect from %s", callloc);
      ALogEvent("Connect from %s", callloc);

      child = fork();

      if (child == -1) {
         syslog(LOG_ERR, "Cannot fork [%d]", errno);
         close(m);
         continue;
      }
      if (child) {
         syslog(LOG_DEBUG, "Forked process %i", child);
         close(m);
      } else {
         close(s);
         Session(m);

         if (fSysLog != NULL)
            fflush(fSysLog);
         fflush(fAdmLog);

         exit(0);
      }
   }
}


void tryname(struct in_addr a, char *m)
{
   struct hostent *h;

   h = gethostbyaddr((char *) &a, sizeof(a), AF_INET);

   if (h)
      strcpy(m, h->h_name);
   else
      strcpy(m, inet_ntoa(a));
}

char      bigbuff[6000];
char      bb2[6000];

#define TIMEOUT 600

int       urgsock;

void Timer(int t)
{
#if defined(DONT_CARE_ABOUT_POSIX_COMPLIANCE)
   struct itimerval tv;

   tv.it_value.tv_sec = timeout;
   tv.it_value.tv_usec = 0;
   tv.it_interval.tv_sec = 0;
   tv.it_interval.tv_usec = 0;
   setitimer(ITIMER_REAL, &tv, NULL);
#else /* ! DONT_CARE_ABOUT_POSIX_COMPLIANCE */
   alarm(t);
#endif /* DONT_CARE_ABOUT_POSIX_COMPLIANCE */
}


void Session(int s)
{
   int       k,
             n;

   struct sigaction sa;

   nLog = 0;


   sa.sa_handler = sig_alarm;
   sigemptyset(&sa.sa_mask);
   sa.sa_flags = 0;
   sigaction(SIGALRM, &sa, NULL);

   sa.sa_handler = sig_urgent;
   sigemptyset(&sa.sa_mask);
   sa.sa_flags = 0;
   sigaction(SIGURG, &sa, NULL);

   urgsock = s;
   fcntl(s, F_SETOWN, getpid());
   n = AllocBuffers(nbuff);

   while (1) {
   		int bp=0;
      if (timeout)
         Timer(timeout);
      		while (1) {
      		k = read(s, bigbuff+bp, sizeof(bigbuff)-bp);
      		bp+=k;
      if (k == 0) {
         Logout();
         ALogEvent("Client closed socket");
         syslog(LOG_NOTICE, "Client closed socket");
         exit(1);
      }
      		if (bigbuff[bp-1]==0) break;          
      		}
      		
/*
 * Now we should clear the timer and prepare for interrupts
 */

      if (timeout)
         Timer(0);

      XServer(bigbuff, bb2);


      if (nLog == -1)
         break;

      write(s, bb2, strlen(bb2) + 1);
   }

   close(s);
}


void sig_alarm(int sig)
{
   syslog(LOG_WARNING, "Timed out");
   ALogEvent("Timed out");
   Logout();
   exit(1);
}

void sig_urgent(int sig)
{
   char      buffer[100];

   recv(urgsock, buffer, sizeof(buffer), 0);
   syslog(LOG_NOTICE, "Client sent interrupt");
   bInt = TRUE;
}

#else /* ! __unix__ */

#define DEFPARAM "corpus.prm"

# if !defined(WIN32)
int main()
{
   _fullpath(param_file, DEFPARAM, sizeof(param_file));
   return 1;
}

# else /* ! WIN32 */
#  define DLL_PROCESS_ATTACH 1
BOOL __stdcall DllMain(void *hInst,
		                 long unsigned ul_reason_for_call,
		                 void *lpReserved)
{
   switch (ul_reason_for_call) {
    case DLL_PROCESS_ATTACH:
//      _fullpath(param_file, DEFPARAM, sizeof(param_file));
      break;

    default:
      break;
   }
   return TRUE;
}

# endif /* WIN32 */

# if defined(WIN32)
__declspec(dllexport) InitServer(const char* p,BOOL(*f) (void))
# else /* ! WIN32 */
__far __pascal __export __loadds InitServer(const char* p,BOOL(__far __pascal * f) (void))
# endif /* WIN32 */
{
//   char      lpath[_MAX_PATH];
   int       n;
   if (p[0]==0) _fullpath(param_file, DEFPARAM, sizeof(param_file));
   else strcpy(param_file,p);

   interrupt_function = f;
	if (!Startup(param_file)) return 0;
//   openlog("sara", 0, 0);
/*   sprintf(lpath, "%ssyslog", accpath);


   if (bSysLog)
      fSysLog = fopen(lpath, "a");
*/
   n = AllocBuffers(nbuff);
   ClearBuffers();
   nLog = 1;
   return 1;
}

#endif /* __unix__ */

int nprmerrs;

BOOL Startup(char* file)
{                   
	char dpath[_MAX_PATH]; nprmerrs=0;
   bInit=CLine(file,&nprmerrs);                                            
   
   strcpy(dpath,etcpath);
   strcat(dpath,corpname);
   strcat(dpath,".dsc");
   ReadDescFile(dpath);
	GetPIndex();   
	if (!bSkipTests && !Tests()) return 0;
   openlog("sara", 0, 0);
	   
   GetBIndex();  
   syslog(LOG_NOTICE,"Opened bibliography index");  
   GetAccels();                                                        
   syslog(LOG_NOTICE,"Initialised accelerators");
   GetDict();                                                      
   syslog(LOG_NOTICE,"Opened dictionary");
   SCReadAll();
   syslog(LOG_NOTICE,"Read subcorpus register");
//   initialise_lexrules(NULL);
   return TRUE;
}

#if defined(SERVER)

char      ans[MAX_SOL + 10];

struct pcmd pc;


void

# if !defined(__unix__)
#  if defined(WIN32)
__declspec(dllexport)
#  else /* ! WIN32 */
__far __pascal __export
#  endif /* WIN32 */
# endif /* __unix__ */

XServer(char *c, char *cc)
{

# if !defined(__unix__)
   _fmode = _O_BINARY;
# endif /* __unix__ */
   syslog(LOG_DEBUG, "Client sent string %s", c);
   ParseString(c, &pc, ctable);
   ExecString(&pc);
   strcpy(cc, ans);   
   syslog(LOG_DEBUG, "Sent reply %.500s", cc);
   return;
}

#else /* ! SERVER */

void Server(void)
{
   char      s[MAX_COMMAND];
   struct pcmd pc;

   ntry = nLog = 0;

   do {
      GetCString(s);
      if (!ParseString(s, &pc, ctable))
         continue;
      ExecString(&pc);
   }

   while (nLog != -1);
}


void GetCString(char *c)
{
   printf("server> ");
   scanf("%[^\n]", c);
   while (getchar() != 10)
      ;
   return;
}

#endif /* SERVER */

#if defined(__unix__)
BOOL TestInterrupt(void)
{
   return bInt;
}

#else /* ! __unix__ */

BOOL TestInterrupt(void)
{
   bInt = bInt || (*interrupt_function) ();
   return bInt;
}

#endif /* __unix__ */

BOOL ParseString(char *s, struct pcmd *pc, struct cmd table[])
{
   char      cd0[10];
   char     *cd = cd0;
   int       i;
   char      c1;

   while (*s == 32)
      s++;

   while (*s != 0 && *s != 32) {
      c1 = *(s++);
      *(cd++) = toupper(c1);
   }

   *cd = 0;

   if (!LookUpCmd(cd0, &(pc->i), table)) {
      SetString("NO COMMAND");
      return FALSE;
   }

   for (i = 0; i < MAX_OPS; i++)
      GetArg(table[pc->i].op[i], &s, &(pc->arg[i]));

   return TRUE;
}


#if defined(SERVER)

void SetString(char *s)
{
   strcpy(ans, s);
   return;
}

#else /* ! SERVER */

void SetString(char *s)
{
   printf("%s\n", s);
   return;
}

#endif /* SERVER */


BOOL LookUpCmd(char *cd, int *i, struct cmd table[])
{
   int       j;

   for (j = 0; j < NCOMMAND; j++)
      if (!strcmp(cd, table[j].name))
	 break;

   *i = j;

   return j < NCOMMAND;
}


void GetArg(int op, char **s, union uarg * arg)
{
   char      t[10];

   switch (op) {
    case OP_USER:
      GetBTS(s, arg->user);
      return;

    case OP_PW:
      GetBTS(s, arg->pw);
      return;

    case OP_NAME:
      GetBTS(s, arg->name);
      return;

    case OP_NUM:
      GetBTS(s, t);
      arg->num = atoi(t);
      return;

    case OP_LNUM:
      GetBTS(s, t);
      arg->lnum = atol(t);
      return;

    case OP_CQL:
      GetRest(s, arg->cql, MAX_CQL);
      return;

    case OP_REGEXP:
      GetBTS(s, arg->regexp);
      return;

    case OP_TABLE:
      GetTable(s, &(arg->num));
      return;

    case OP_FILTER:
      GetRest(s, arg->filter,MAX_FILTER);
      return;

    case OP_RANGE:
      GetRange(s, &(arg->range));
      return;

    default:
      return;
   }
}


void GetBTS(char **s, char *c)
{
   while (isspace(**s))
      (*s)++;

   if (**s == '\"') {           /* " for emacs */
      (*s)++;
      while (**s != 0 && **s != '\"') /* " for emacs */
         *(c++) = *((*s)++);

      *c = 0;

      if (**s != 0)
         (*s)++;

      return;
   }

   while (**s != 0 && !isspace(**s))
      *(c++) = *((*s)++);

   *c = 0;

   return;
}


void GetRest(char **s, char *c,int max)
{
   while (**s > 0 && isspace(**s))
      (*s)++;

   while (**s != 0) {
      *(c++) = *((*s)++);
      max--;
      if (max <=0) break;
   }   

   *c = 0;

   return;
}


void GetRange(char **s, struct ran *r)
{
   char      c0[10];
   char     *c = c0;

   while (isspace(**s))
      (*s)++;

   while (**s != 0 && !isspace(**s) && **s != ';')
      *(c++) = *((*s)++);

   r->min = atoi(c);
   r->size = 1;

   if (**s != ';')
      return;

   c = c0;

   while (**s != 0 && !isspace(**s))
      *(c++) = *((*s)++);

   r->size = atoi(c);

   return;
}

#define MAXSELTAB 1000

static int       seltab[MAXSELTAB];

void GetTable(char **s, int *i0)
{
   int       i = 0;
   int       j = 0;
   BOOL      bInt = FALSE;

   while (**s != 0) {
      if (**s == ' ' && bInt && i<MAXSELTAB) {
         seltab[i++] = j;
         bInt = FALSE;

         j = 0;

      } else if (isdigit(**s)) {
         j = 10 * j + (**s - '0');
         bInt = TRUE;
      }
      (*s)++;
   }

   if (bInt && i<MAXSELTAB)
      seltab[i++] = j;

   *i0 = i;

   return;
}

struct pcmd pf;


void ExecString(struct pcmd *pc)
{
   int         i;
   int         n;
   sara_long   ln;

#if !defined(SERVER)
   printf("%i\n", pc->i);
#endif /* SERVER */

   for (i = 0; i < MAX_OPS; i++)
      parg(ctable[pc->i].op[i], &(pc->arg[i]));

   if (nLog != 1 && pc->i > 1 && pc->i != 32) {
      SetString("NO LOGIN");
      return;
   }

   switch (pc->i) {
    case 0:				 /* LOGOUT */
      Logout();
      SetString("");        
      CleanUp();

      break;

    case 1:				 /* LOG */
      Log(pc->arg[0].user, pc->arg[1].pw);
      break;

    case 2:				 /* SAVE */
      {
         int       i;
         BOOL      b = (pc->arg[0].num != 0);

         for (i = 0; i < NFILES; i++)
            if (qtable[i].bUse &&
                !strcmp(qtable[i].name, pc->arg[1].name))
               break;

         qtable[i].bSave = b;
         SetString("OK");
         break;
      }

    case 3:				 /* DIR */
      /* FALLTHROUGH */
      
    case 4:				 /* REMOVE */
      FreeQPath(pc->arg[0].name);
      SetString("OK");
      break;

    case 5:				 /* PURGE */
      /* FALLTHROUGH */

    case 6:				 /* SOLVE */

		if (pc->arg[1].cql[0]==0) {
			char s[40];
			ln=Solve0(pc->arg[0].name,&i);
			sprintf(s, "OK %ld %d", ln, i);
			SetString(s);
			break;
		}

      if (!NewQName(pc->arg[0].name)) {
         SetString("NO FILES");
         break;
      }
      ln = Solve(pc->arg[0].name, pc->arg[1].cql, &i);

      if (ln <= 0)
         remove(GetQTPath(pc->arg[0].name));
      else {
         remove(GetQPath(pc->arg[0].name));
         rename(GetQTPath(pc->arg[0].name), GetQPath(pc->arg[0].name));
      }

      if (ln == -4)
         SetString("NO STREAMS");
      else if (ln == -3)
         SetString("NO ABORT");

      else if (ln == -2)
         SetString("NO SPACE");

      else if (ln == -1)
         SetString("NO SYNTAX");

      else if (ln == 0)
         SetString("NO 0");

      else {
         char      s[40];

         sprintf(s, "OK %ld %d", ln, i);
         SetString(s);
      }

      break;
      
    case 7:				 /* SUBCORPUS */
      {
         char      s[20]; double m, sd; long nw, nu;

         if (strcmp(pc->arg[0].name, corpname))
            break;
			m=0.0;
			sd=0.0;
			nw=dicthead.nw;
			nu=dicthead.nu;
         sprintf(s, "OK %d %f %f %ld %ld", idxhead.n, m, sd, nw, nu);
         SetString(s);
         break;
      }

   case 8:				 /* WORDLIST */
      /* This is superseded by the {} syntax in a query	*/

      n = WoL1(pc->arg[1].cql);

      if (n == -2)
         SetString("NO TOOMANY");

      else if (n == -1)
         SetString("NO SYNTAX");

      else if (n == 0)
         SetString("NO 0");

      else {
         char      s[40];

         sprintf(s, "OK %d", n);
         SetString(s);
      }
      break;

   case 9:				 /* SQTABLE */
      {
         int       i = pc->arg[2].num;
         char      s[10];
         int       j;

         if (!NewQName(pc->arg[1].name)) {
            SetString("NO FILES");
            break;
         }
         SQTable(pc->arg[0].name,pc->arg[1].name, i, &j);

         if (i > 0)
            sprintf(s, "OK %d %d", i, j);
         else
            strcpy(s, "NO 0");

         SetString(s);
         break;
      }

    case 10:				 /* FILTER */
      if (pc->arg[1].filter[0] == 0) {
         ClearFilters(pc->arg[0].name);
         break;
      }
      if (!ParseString(pc->arg[1].filter, &pf, ftable))
         break;

      AddFilter(pc->arg[0].name, &pf);
      break;

    case 11:				 /* SORTFILTER */
      /* FALLTHROUGH */
      
    case 12:				 /* SORT */
      /* FALLTHROUGH */

    case 13:				 /* GET */
      {
         char       *s;
         char        s1[MAX_SOL];
         int         pos;
	      int         len;
         sara_long   code;

         char         sno[20];
         char        id[20];

         bPOS = bSLen = bText = FALSE;
		 strcpy(sno,"?");
         s = GetSol(pc->arg[0].name, pc->arg[1].num, &pos, &len,
                    &code, pc->arg[2].cql, sno);
         
         if (code==-1) {
         	SetString("NO SOL");
         	break;
         }	

         if (s == NULL) {
         	sprintf(s1,"OK -1 -1 0 0 HDR Text not available",id);
            SetString(s1);
            break;
         }
         Reformat(pc->arg[0].name, s, &pos, &len);
		 {
			 char path[_MAX_PATH];
			 GetDocPath(code,id,path);
		 }
         if (len > (int) strlen(s) - pos && (int) strlen(s) - pos >= 0)
            len = strlen(s) - pos;

         sprintf(s1, "OK %s %s %d %d %s %s", id, sno, pos, len,
			 bNoPos?"HDR":postab[nPOSx%10000].name, s);
         SetString(s1);
         break;
      }

    case 14:				 /* CUT */
      break;

    case 15:				 /* CHAR */
      break;

    case 16:				 /* GETPOS */
      {
         int       nsig[20],
                   i,
                   n;
         char      s[100];

         n = PosList(pc->arg[0].name, nsig);

         sprintf(s, "OK %d", n);

         for (i = 0; i < n; i++) {
            strcat(s, " ");
            strcat(s, postab[nsig[i]-(c6?C6OFFSET:0)].name); 
         }
         SetString(s);
         break;
      }

    case 17:				 /* GETDTD */
      break;

    case 18:				 /* MAXLENGTH */
      {
         char      s[20];
         int       n = pc->arg[0].num;

         if (n > MAXSENTENCE)
            n = MAXSENTENCE;

         max_sent = n;
         sprintf(s, "OK %d", n);
         SetString(s);
         break;
      }

   case 19:				 /* GETSC */
      {
         char      s[20],
                   s1[20];
         char      path[_MAX_PATH] = "";
         FILE* f; BOOL b;
         
         if (strcmp(pc->arg[0].name, corpname))
            {
            SetString("NO");
            break;
         }   

         GetDocPath(pc->arg[1].num, s, path);
         f=fopen(path,"r");
         b=(f!=NULL);
         if (b) fclose(f);
         sprintf(s1, "OK %s %d", s,b);
         SetString(s1);
         break;
      }

   case 20:				 /* GETHEAD */
      {
         char        text[10];
         char       *s;
         char        s1[MAX_SOL];
         BOOL        bj;
         sara_long   jump;
         int         j = 0;
         int         jj = 0;
         sara_long   pos;
         sara_long   newpos;
         int         depth;
         int         newdepth;

         strcpy(text, pc->arg[0].name);
         newpos = pc->arg[1].lnum;
         newdepth = pc->arg[2].num;

         do {
            pos = newpos;
            depth = newdepth;
            s = GetHeadElt(text, pos, depth, &newpos, &newdepth,
                           &bj, &jump);
         } while (newdepth > 0 && AllBlank(s));

         if (s==NULL) {
         	SetString("NO TEXT");
         	break;
         }
         Trim(s, &j, &jj);
         sprintf(s1, "OK %li %li %i %i %s", newpos, jump, newdepth, bj, s);
         SetString(s1);
         break;
      }

   case 21:				 /* LOOKUP */
      {
         sara_long   x1;
         sara_long   x2;
         char        s[40];
			LowerCase(pc->arg[0].name);
         if (GetLims(fDict, 0, ndix - 1, pc->arg[0].name, &x1, &x2)) {
            dbase = x1;
            sprintf(s, "OK %d", x2 - x1 + 1);
         } else
            strcpy(s, "NO 0");

         SetString(s);
         break;
      }

   case 22:				 /* DMATCH */
      {
         char        s1[80] = "";
         char        s2[80] = "";
         char        s[200];    /* used to be 100, but 80 + 80 > 100 */
         sara_long   freq, forms;
         int         i = 0;
         int         j = 0;
         int         n = pc->arg[0].num;
         sara_long   k;

         k = n + dbase;

         GetDictEntry(fDict, k, s1, sizeof(s1), &freq, &forms);
         strcpy(s2, s1);
         TranslateChars(s2, &i, &j); if (s2[0]==0) strcat(s2, " ");
	if (!SCRegistered()) freq=QuickFrequency(s1);
         sprintf(s, "OK %li %s{%s} %li", freq, s1, s2,forms);

         SetString(s);
         break;
      }

   case 23:				 /* QNAME */
      {
         char      s[40];
         char      n[20];

         do
            sprintf(n, "hq%06d", iq++);
         while (Exists(n));

         Dummy(n);
         sprintf(s, "OK %s", n);
         SetString(s);
         break;
      }

   case 24:				 /* OPEN */
      {
         int       k = OpenQPath(pc->arg[0].name);
         char      s[40];

         if (k > 0)
            sprintf(s, "OK %d", k);
         else
            strcpy(s, "NO NOFILE");

         SetString(s);
         break;
      }

   case 25:				 /* CSCORE */
      {
         char        s[40]; int t;
         sara_long   l = colloc_score(pc->arg[0].name, pc->arg[2].cql,
                                      pc->arg[1].num,pc->arg[1].num,&t);

         if (l < 0)
            strcpy(s, "NO SYNTAX");
         else
            sprintf(s, "OK %ld %d", l, t);

         SetString(s);
         break;
      }

   case 26:				 /* LOC */
      {
         char              s[40];
         FILE             *f;
         struct locent1    l;
         char             *name;

         name = GetQPath(pc->arg[0].name);
         if ((f = fopen(name, "r")) == NULL) {
            syslog(LOG_ERR, "Can't open %s\n", name);
            strcpy(s, "NO FILE");
            break;
         } else {
            fseek(f, (off_t)(pc->arg[1].num * sizeof(struct locent1)),
                  SEEK_SET);
            fread(&l, sizeof(l), 1, f);
            fclose(f);

            sprintf(s, "OK %ld %ld %d", l.lcode, l.pos, l.lw);
         }
         SetString(s);
         break;
      }

   case 27:				 /* GETHEAD2 */
      {
         char        text[10];
         char       *s;
         char        s1[MAX_SOL] = "";
         sara_long   posx;
         int         lenx;
         BOOL        bj = 0;
         sara_long   jump = 0;
         sara_long   pos;
         sara_long   newpos;
         int         depth;
         int         newdepth;

         strcpy(text, pc->arg[0].name);
         newpos = pc->arg[1].lnum;
         newdepth = 100;
         posx = pc->arg[2].lnum;
         lenx = pc->arg[3].num;

         do {
            pos = newpos;
            depth = newdepth;
            s = GetHeadElt(text, pos, depth, &newpos, &newdepth, &bj, &jump);
            if (s==NULL) break;
         }                              
         while (newdepth > 0 && AllBlank(s))
            ;
         
		if (s==NULL) {SetString("NO TEXT"); break;}
         posx -= pos;
         Trim(s, (int *) &posx, &lenx);
         
         sprintf(s1, "OK %li %i", posx, lenx);
         SetString(s1);
         
         break;
      }
      
    case 28:				 /* GETCHEAD */
      break;
      
    case 29:				 /* GET1SOL */
      {
         char      text[10];
         char      element[10];
         char      att[10];
         char      s[1000];

         strcpy(s, "OK ");
         strcpy(text, pc->arg[0].name);
         strcpy(element, pc->arg[1].name);
         strcpy(att, pc->arg[2].name);

         Get1Sol(FindDoc(text), element, att, s + 3);
         SetString(s);

         break;
      }
      
    case 30:				 /* MOTD */
      {
         char      s[2000];
         int       k;
         FILE     *f;
         char      mpath[_MAX_PATH];

         sprintf(mpath, "%smotd.txt", etcpath);

         f = fopen(mpath, "r");
         if (f == NULL) {
            syslog(LOG_WARNING, "Can't open MOTD %s\n", mpath);
            SetString("NO");
            break;
         }
         strcpy(s, "OK ");

         k = fread(s + 3, 1, sizeof(s) - 3, f);
         s[k + 3] = 0;

         fclose(f);
         SetString(s);
         break;
      }
      
    case 31:				 /* PWD */
      {      
#if defined(__unix__)      
         if (SetPassword(calluser, pc->arg[0].pw, pc->arg[1].pw))
            SetString("OK");
         else
            SetString("NO");
         break;    
#endif         
      }
      
    case 32:				 /* INFO */
      /*
       * Format is as follows:
       * 
       * Client sends
       *    INFO cs
       * where
       *    cs is the character set ID
       * 
       * Server sends
       *    OK t hdv sv cv nm k
       * where
       *    t is the timeout
       *    hdv is the header version
       *    sv is the server version
       *    cv is the lowest acceptable client version 
       *    nm is the corpus name
	   *    k is the number of registered subcorpora (excluding all)
       *
       */
      {
         char      s[30];
         char      t[30];

         strcpy(t, pc->arg[0].name);
//         LoadChars(t);
         sprintf(s, "OK %d %d %d 931 %s %d", timeout, hdv, VER,corpname,SCCount()-1);

         SetString(s);
         break;
      }
      
    case 33:				 /* RLOOKUP */
      {
         int       n;
         char      s[30];

         n = WoL1(pc->arg[0].cql);
         if (n == -2){
			 sprintf(s,"NO TOOMANY %d",MAXRE);
			 SetString(s);
		 }
         else if (n == 0)
            SetString("NO 0");
         else {
            sprintf(s, "OK %i", n);
            SetString(s);
         }

         break;
      }
      
    case 34:				 /* RGET */
      {
         int         n;
         char        s[80];
         sara_long   freq = 0;
		 sara_long forms=0;
         char        wd[80] = "";
         char        wd1[80] = "";
         int         i = 0;
         int         j = 0;

         n = pc->arg[0].num;

         WoL2(n, wd, sizeof(wd), &freq, &forms);
		 	if (isc) freq=QuickFrequency(wd);

         strcpy(wd1, wd);
         TranslateChars(wd1, &i, &j); if (wd1[0]==0) strcpy(wd1," ");

         sprintf(s, "OK %li %s{%s} %li", freq, wd, wd1,forms);
         SetString(s);
         break;
      }

    case 35:				 /* BIB */
      {
         char      s[1000];
         int       ty = 0;
         int       n = 0;

         if (GetBibHeader(pc->arg[0].name, &ty, &n)) {
            sprintf(s, "OK %d %d", ty, n);
            SetString(s);
         } else
            SetString("NO BIB");
         break;
      }
      
    case 36:				 /* BIBITEM */
      {
         char      s[1000];

         strcpy(s, "OK ");
         if (GetBibItem(pc->arg[0].name, pc->arg[1].num, s + 3))
            SetString(s);
         else
            SetString("NO BIB");
         break;
      }
      
    case 37:				 /* TRACE */
      bTrace = pc->arg[0].num;
      SetString("OK");
      break;
      
    case 38:				 /* DOWNLOAD */
      {
         char      path[_MAX_PATH] = "";

         if (!strcmp(pc->arg[0].name, "header")) {
         	char buffer[_MAX_PATH];
			strcpy(buffer,etcpath);
			strcat(buffer,corpname);
			strcat(buffer,".dsc");
 			fDownload=fopen(buffer,"r");
         } else {
            strcpy(path, etcpath);
            strcat(path, pc->arg[0].name);
            strcat(path, ".txt");
            if ((fDownload = fopen(path, "r")) == NULL) {
               syslog(LOG_ERR, "Can't open %s for download\n", path);
               SetString("NO FILE");
               return; 
               }
         }
         SetString("OK");
         return;
      }

    case 39:				 /* LINE */
      {
   		int i; char *s, *s1;
   		if (feof(fDownload))  {
   			fclose(fDownload);
   			SetString("NO MORE");
   			return;
   		}	
   		strcpy(ans,"OK ");
   		i=fread(ans+3,1,MAXSENTENCE,fDownload);
   		ans[i+3]=0;
		s=s1=ans+3;
		while (*s) {
			if (*s!='\r') *s1++=*s; 
			s++;
		}
		*s1=0;
   		
   /*      char      buffer[100];
         int       i;

         strcpy(buffer, "OK ");
         if (bDTD) {
            if (DTDelt(nDTD++, buffer + 3))
               SetString(buffer);
            else
               SetString("NO MORE");
            return;
         }
   
         if (!fgets(buffer + 3, sizeof(buffer) - 3, fDownload)) {
            SetString("NO MORE");
            fclose(fDownload);
            return;
         }
         for (i = 0; i < (int) strlen(buffer); i++)
            if (buffer[i] == '\n' || buffer[i] == '\r')
               buffer[i] = 0;
         SetString(buffer);
   */      
         return;
      }            
   
    case 40:				 /* THIN */
      {
         char      s[40];
         int       i1 = 0;
         int       i2 = 0;

         if (!NewQName(pc->arg[1].name)) {
            SetString("NO FILES");
            break;
         }
         Thin(pc->arg[0].name, pc->arg[1].name, pc->arg[2].num,
              pc->arg[3].num, pc->arg[4].num, &i1, &i2);
         sprintf(s, "OK %d %d", i1, i2);
         SetString(s);
         break;
      }
	case 41:     /* WEB */
		{
		char s[_MAX_PATH+10];
		if (website[0]==0) SetString("NO"); else {
			sprintf(s,"OK %s",website);
			SetString(s);
		}
		break;
		}
	case 42:     /* SCDEF */
		{
			char s[100];
			if (NewSubCorpus(pc->arg[0].name))
				sprintf(s,"OK %d",MAXSELTAB); 
			else strcpy(s,"NO");
				SetString(s);
			break;
		}
	case 43:	 /* SCADD */
		{
			AddSubCorpus(pc->arg[0].name,pc->arg[1].num,seltab);
			SetString("OK");
			break;
		}
	case 44:	 /* SCSET */
		{
			if (SetSubCorpus(pc->arg[0].name))
				SetString("OK");
			else
				SetString("NO");
			break;
		}
	case 45: /* FTAB */
		{
			char s[100];
			int n;
			n=FTable(pc->arg[0].num,pc->arg[1].num,pc->arg[2].num,pc->arg[3].cql);
			sprintf(s,"OK %d",n);
			SetString(s);
			break;
		}
	case 46: /* FENTRY */
		{
			char s[1000]; char t[MAXWORD]; int k, k1;
			FEntry(pc->arg[0].num,t,&k, &k1);
			sprintf(s,"OK {%s} %d %d",t,k, k1);
			SetString(s);
			break;
		}
	case 47: /* FFREE */
		FFree();
		SetString("OK");
		break;
	case 48: /* SCSAVE */
		{ char p[_MAX_PATH];
			char s[_MAX_PATH+3];
			if (SCSave(p))
			   sprintf(s,"OK %s",p);
			else sprintf(s,"NO");
			   SetString(s);
			break;
		}
	case 49:  /* SCLIST */
		{
			char s[1000];
			sprintf(s,"OK %s",SCList(pc->arg[0].num));
			SetString(s);
			break;
		}
	case 50:  /* SCENUM */
		{
			char s[1000];
			sprintf(s,"OK %s",SCEnum(pc->arg[0].name,pc->arg[1].num,pc->arg[2].num));
			SetString(s);
			break;
		}
	case 51: /* CTAB */
		{
			char s[1000];
			int n=MakeCTab(pc->arg[0].name,pc->arg[1].num,pc->arg[2].num,pc->arg[3].cql);
			if (n >= 0) sprintf(s,"OK %d",n); else strcpy(s,"NO");
			SetString(s);
			break;
		}
	case 52: /* CTABENTRY */
		{
			char s[1000]; char* w; int f; double sc;
			CTab(pc->arg[0].num,&w,&f,&sc);
			sprintf(s,"OK {%s} %d %f",w,f,sc);
			SetString(s);
			break;
		}
	case 53: /* ACSCORE */
		{
			char s[1000]; double sc; long d;
			ACScore(pc->arg[0].name,pc->arg[1].num,pc->arg[2].num, pc->arg[3].num, pc->arg[4].cql,
				&d,&sc);
			sprintf(s,"OK %d %f",d,sc);
			SetString(s);
			break;
		}
	case 54: /* CTABTHIN */
		{
			char s[1000]; int i1=0; int i2=0;
			if (!NewQName(pc->arg[1].name)) {
				SetString("NO FILES");
				break;
			}
			CTabThin(pc->arg[0].name,pc->arg[1].name,
				pc->arg[2].num,pc->arg[3].num,pc->arg[4].cql,
				&i1,&i2);
			sprintf(s,"OK %d %d",i1, i2);
			SetString(s);
			break;
		}
	case 55: /* CTABOPTIONS */
		CTabOptions(pc->arg[0].num,pc->arg[1].num,pc->arg[2].name,pc->arg[3].num);
		SetString("OK");
		break;
	case 56: /* FORM */
		{
			int i=pc->arg[0].num; char s[1000]; int f;
			char form[MAXWORD];
			if (Form(pc->arg[1].name,i,form,&f))
				sprintf(s,"OK {%s} %d",form,f);
			else 
				strcpy(s,"NO");
			SetString(s);
			break;
		}
	case 57: /* LEMMINFO */
		{
			char s[1000];
			sprintf(s,"OK %s",LemmInfo());
			SetString(s);
			break;
		}
	case 58: /* LEMMALL */
		{
			char s[1000]; int i;
			strcpy(s,"OK ");
			for (i=0;i<LemmaNSchemes();i++) {
				strcat(s,LemmaName(i));
				strcat(s," ");
			}
			SetString(s);
			break;
		}
	case 59: /* LEMMSEL */
		{
			int i;
			for (i=0;i<LemmaNSchemes();i++) {
				if (!strcmp(pc->arg[0].name,LemmaName(i))) break;
			}
			if (i==LemmaNSchemes())
				SetString("NO");
			else {
				LemmaSetScheme(i,TRUE);
				SetString("OK");
			}
			break;
		}
    case 60:				 /* GETSOL */
      {
         char       *s;
         char        s1[MAX_SOL];
         int         pos;
	      int         len;
         sara_long   code;

         char         sno[20];

         bPOS = bSLen = bText = FALSE;
		 strcpy(sno,"?");
         s = GetSol(pc->arg[0].name, pc->arg[1].num, &pos, &len,
                    &code, pc->arg[2].cql, sno);
         
         if (code==-1) {
         	SetString("NO SOL");
         	break;
         }	

         if (s == NULL) {
         	sprintf(s1,"OK -1 -1 0 0 HDR Text not available");
            SetString(s1);
            break;
         }
         Reformat(pc->arg[0].name, s, &pos, &len);
         if (len > (int) strlen(s) - pos && (int) strlen(s) - pos >= 0)
            len = strlen(s) - pos;
         sprintf(s1, "OK %d %s %d %d %s %s", code, sno, pos, len,
			 bNoPos?"HDR":postab[nPOSx%10000].name, s);
         SetString(s1);
         break;
      }
	case 61: /* CTABFREE */
		CTabFree();
		SetString("OK");
		break;
	case 62: /* LEMMDESC */
		{
			int i;
			char s[1000];
			for (i=0;i<LemmaNSchemes();i++) {
				if (!strcmp(pc->arg[0].name,LemmaName(i))) break;
			}
			if (i==LemmaNSchemes())
				SetString("NO");
			else {
				char* d;
				LemmaDescribe(i,&d);
				sprintf(s,"OK %s",d);
				SetString(s);
			}
			break;
		}
	case 63: /* LEMMTEST */
		{
			int i;
			for (i=0;i<LemmaNSchemes();i++) {
				if (!strcmp(pc->arg[0].name,LemmaName(i))) break;
			}
			if (i==LemmaNSchemes())
				SetString("NO");
			else {
				char s[1000]="OK ";
				if (LemmaTest(i,pc->arg[3].num,pc->arg[1].name,pc->arg[2].name,s+3))
					SetString(s);
				else
					SetString("NO");
			}
			break;
		}









   }
   return;
}

BOOL AllBlank(char *s)
{
   while (*s != 0) {
      if (!isspace(*s))
         return FALSE;
      s++;
   }

   return TRUE;
}


void parg(int op, union uarg * a)
{
   switch (op) {
    case OP_USER:               /* FALLTHROUGH */
    case OP_PW:                 /* FALLTHROUGH */
    case OP_NAME:               /* FALLTHROUGH */
    case OP_CQL:                /* FALLTHROUGH */
    case OP_REGEXP:             /* FALLTHROUGH */
    case OP_FILTER:
      return;

    case OP_NUM:
      return;

    case OP_TABLE:
      return;

    case OP_RANGE:
      return;

    default:
      return;
   }
}

void trace(char *ty, char *format,...)
{
   time_t   t;
   char     tt[30];
   char     buf[4096];

   va_list(args);
   time(&t);
   strcpy(tt, ctime(&t));
   tt[strlen(tt) - 1] = 0;
   va_start(args, format);
   vsprintf(buf, format, args);
   syslog(LOG_NOTICE, "%s", buf);
}

#if !defined(__unix__)
static char *logID = 0;
#define LOGHOSTLEN   128
static char logHost[LOGHOSTLEN+1];

void openlog(char *id, int flags, int facility)
{
   char  ap[_MAX_PATH];
   char *pch;
   
      
#if defined(__unix__)
   if (logID == 0) {
      sprintf(ap, "%ssyslog", accpath);
      fSysLog = fopen(ap, "w");
      logID = strdup(logID);
      gethostname(logHost, LOGHOSTLEN);
      logHost[LOGHOSTLEN] = '\0';
#else   /* __unix__ */
	if (bSysLog) {
		sprintf(ap,"%ssyslog",accpath);
		fSysLog=fopen(ap,"w");
		if (fSysLog==NULL) bSysLog=FALSE;
	  logHost[0] = '\0';
#endif	/* __unix__ */  	      
      if ((pch = strchr(logHost, '.')) != 0)
         *pch = '\0';
   }
}

void syslog(int level, char *format,...)
{
   time_t    t;
   char      tt[30];

   va_list(args);
   if (fSysLog == NULL)
      return;
   time(&t);
#if !defined(__unix__)        
  	strcpy(tt,ctime(&t));
	tt[strlen(tt)-1]=0;
#else
   strftime(tt, 30, "%b %d %T", localtime(&t));
#endif
   fprintf(fSysLog, "%s %s[%d]: ", tt, logHost, getpid());
   va_start(args, format);
   vfprintf(fSysLog, format, args);
   fprintf(fSysLog, "\n");
   fflush(fSysLog);
}
#endif /* ! __unix__ */

#ifdef __unix__
void ALogEvent(char *format,...)
{
   time_t    t;
   char      tt[30];

   va_list(args);
   if (fAdmLog == NULL)
      return;
   time(&t);
  	strcpy(tt,ctime(&t));
	tt[strlen(tt)-1]=0;
   fprintf(fAdmLog, "%s local[%d]: ", tt, getpid());
   va_start(args, format);
   vfprintf(fAdmLog, format, args);
   fprintf(fAdmLog, "\n");
   fflush(fAdmLog);
}
#endif

void Log(char *user, char *pw)
{              
#if defined(__unix__)
   if (!LogUser(user, pw)) {
      SetString("NO BADLOG");
      /*
       * Took out the bits that printed the user name and password.
       *
       * People often accidentally type their password when prompted for
       * their username so it's best not to log them.
       */
 
      ALogEvent("Unsuccessful attempt at log from %s (attempt %d)",
                callloc, ntry);
      syslog(LOG_WARNING, "Unsuccessful attempt at log from %s (attempt %d)",
               callloc, ntry);
      ntry++;

      if (ntry >= MAXTRY)
         nLog = -1;
   } else {
      SetString("OK (C) BNC Consortium, 1995");
      ALogEvent("%s logged on from %s", user, callloc);
      syslog(LOG_NOTICE, "%s logged on from %s", user, callloc);

      strcpy(calluser, user);
      nLog = 1;
      bServer = TRUE;
   }

   return;  
#endif   
}


void Logout(void)
{
   int       i;

   bServer = FALSE;

   for (i = 0; i < NFILES; i++)
      if (qtable[i].bUse && !qtable[i].bSave)
         remove(qtable[i].path);

   if (nLog == 1) {
      syslog(LOG_NOTICE, "%s logged out", calluser);
#ifdef __unix__
      ALogEvent("%s logged out", calluser);
#endif      
   }
   nLog = -1;

   if (fSysLog != NULL)
      fflush(fSysLog);

   return;
}


char *GetSol(char *n, int k, int *pos, int *len, sara_long *code,
             char *scope, char *sno)
{
   FILE             *f;
   int               jj;
   struct locent1    l;
   char             *name;
	*code=-1;
   name = GetQPath(n);
   if ((f = fopen(name, "r")) == NULL) {
      syslog(LOG_ERR, "Couldn't open %s\n", name);
      return NULL;
   }
   fseek(f, (off_t)(k * sizeof(struct locent1)), SEEK_SET);
   jj = fread(&l, sizeof(l), 1, f);

   /*
    * locent1s are written by the server, so we don't need to
    * fix them!
    */
   /* fix_locent1(&l, 1); */
   fclose(f);

   if (jj == 0)
      return NULL;

   *code = l.lcode;
   return GetRef(l.lcode, l.pos, l.word, l.lw, l.lw0, pos, len, scope, sno);
}


void SQTable(char *n, char* nn, int nsols, int *nt)
{
   FILE             *f;
   FILE             *f1;
   int               i;
   sara_ulong        lc = 0xf000000;
   struct locent1    l;
   char             *name;

   *nt = 0;
   name = GetQPath(n);
   if ((f = fopen(name, "r")) == NULL) {
      syslog(LOG_ERR, "Can't open %s\n", name);
      return;
   }
   name=GetQPath(nn);

   if ((f1 = fopen(name, "w")) == NULL) {
      syslog(LOG_ERR, "Can't open %s\n", name);
      return;
   }

   for (i = 0; i < nsols; i++) {
      fseek(f, (off_t)(seltab[i] * sizeof(struct locent1)), SEEK_SET);
      fread(&l, sizeof(l), 1, f);
      fwrite(&l, sizeof(l), 1, f1);
      if (lc != l.lcode) {
         lc = l.lcode;
         (*nt)++;
      }
   }
   fclose(f);
   fclose(f1);
   return;
}

void Thin(char *n, char *n1, int how, int ns, int seed, int *k, int *nt)
{
   switch (how) {
    case 0:
      ThinInit(n, n1, ns, nt);
      *k = ns;
      return;

    case 1:
      srand(seed);
      ThinRand(n, n1, ns, nt);
      *k = ns;
      return;

    case 2:
      OnePerText(n, n1, nt);
      *k = *nt;
      return;
   }
}


BOOL isin(sara_long i, sara_long *sel, int nsel)
{
   int       j;

   for (j = 0; j < nsel; j++)
      if (sel[j] == i)
         return TRUE;
   return FALSE;
}

#if !defined(RAND_MAX)
# define RAND_MAX 0x80000000
#endif /* ! RAND_MAX */


sara_ulong r(sara_long n)
{
   return (sara_ulong) ((((double) n) * rand()) / RAND_MAX);
}


void ThinRand(char *n, char *n1, int ns, int *nt)
{
   sara_long        *sel;
   FILE             *f;
   FILE             *f1;
   sara_ulong        lc = 0xf000000;
   struct locent1    l;
   sara_long         ntot;
   int               nsel;
   int               i;
   char             *name;

   name = GetQPath(n);
   if ((f = fopen(name, "r")) == NULL) {
      syslog(LOG_ERR, "Can't open %s\n", name);
      return;
   }
   
   fseek(f, (off_t)0, SEEK_END);
   ntot = ftell(f) / sizeof(l);
   rewind(f);
   if ((sel = malloc(ns * sizeof(sara_long))) == NULL) {
      syslog(LOG_ERR, "malloc() failed");
#if defined(__unix__)
      exit(1);       
#else
      abort();
#endif	        
   }
   nsel = 0;

   for (i = 0; i < ns; i++) {
      sara_long      k;
      do
         k = r(ntot);
      while (isin(k, sel, nsel));
      sel[nsel++] = k;
   }
   *nt = 0;

   name = GetQPath(n1);
   
   if ((f1 = fopen(name, "w")) == NULL) {
      syslog(LOG_ERR, "Can't open %s\n", name);
   } else {
      for (i = 0; i < ntot; i++) {
         fread(&l, sizeof(l), 1, f);
         if (!isin(i, sel, nsel))
            continue;
         fwrite(&l, sizeof(l), 1, f1);
         if (lc != l.lcode)
            (*nt)++;
         lc = l.lcode;
      }
      fclose(f1);
   }
   fclose(f);
   free(sel);
   return;
}

void ThinInit(char *n, char *n1, int ns, int *nt)
{
   FILE             *f;
   FILE             *f1;
   int               i;
   sara_ulong        lc = 0xf000000;
   struct locent1    l;
   char             *name;
   *nt = 0;

   name = GetQPath(n);
   if ((f = fopen(name, "r")) == NULL) {
      syslog(LOG_ERR, "Can't open %s\n", name);
      return;
   }
   
   name = GetQPath(n1);
   if ((f1 = fopen(name, "w")) == NULL) {
      syslog(LOG_ERR, "Can't open %s\n", name);
   } else {
      for (i = 0; i < ns; i++) {
         fread(&l, sizeof(l), 1, f);
         fwrite(&l, sizeof(l), 1, f1);
         if (lc != l.lcode)
            (*nt)++;
         lc = l.lcode;
      }
      fclose(f1);
   }
   fclose(f);
   return;
}

void OnePerText(char *n, char *n1, int *nt)
{
   FILE             *f;
   FILE             *f1;
   sara_ulong        lc = 0xf000000;
   struct locent1    l;
   char             *name;

   *nt = 0;

   name = GetQPath(n);
   if ((f = fopen(name, "r")) == NULL) {
      syslog(LOG_ERR, "Can't open %s\n");
      return;
   }

   name = GetQPath(n1);
   if ((f1 = fopen(name, "w")) == NULL) {
      syslog(LOG_ERR, "Can't open %s\n");
   } else {
      while (fread(&l, sizeof(l), 1, f) == 1) {
         if (lc == l.lcode)
            continue;
         fwrite(&l, sizeof(l), 1, f1);
         lc = l.lcode;
         (*nt)++;
      }
      fclose(f1);
   }
   fclose(f);
   return;
}

#define MAXTAG 1000

BOOL GetAttrib(sara_ulong code,sara_ulong pos,char* elt,char* att,char* val)
{
	sara_long ls, lt;
	char id[MAXNAME + 1] = "";
	char path[_MAX_PATH];
	FILE* fText;
	char psent[MAXTAG+1];
	char* s;

    ls = LastSGMLp(pos, code, elt, ID_TAGO);
	if (ls==0) return FALSE;
	GetDocPath(code,id,path);
    if ((fText = fopen(path, "r")) == NULL) 
      return 0;
    fseek(fText, (off_t)ls, SEEK_SET);
    lt = fread(psent, 1, MAXTAG, fText);
	psent[lt]=0;
	fclose(fText);
	s=psent;
	while (TRUE) {
		s=strstr(s,att);
		if (s==NULL) return FALSE;
		s+=strlen(att);
		while (isspace(*s)) s++;
		if (*s=='=') {
			s++;
			while (isspace(*s)) s++;
			if (*s=='\"') {
				s++;
				while (*s != '\"') 
					*val++=*s++;
			} else {
				while (isalnum(*s))
					*val++=*s++;
			}
			*val=0;
			return TRUE;
		}
	}
}

int MeasureTag(char* s)
{
	char* s0=s; char c; BOOL bQ=FALSE;
	while (c=*s++) {
		if (c=='>' && !bQ) break; 
		else if (c=='\"') bQ=!bQ;
	}
	return s-s0;
}

int MeasureChar(char* s,int *x)
{
	int c; char ss[100]; char* sss=ss;
	int k=IsChar(s,&c);
	if (k==0) {k=IsSpec(s,&sss);c=-1;}
	*x=k;
	if (s[k-1] != ';') (*x)++;
	if (c>=0 && c<256 && c != '&') *x=1;
	return k;
}

void AdjustLength(char* text,int* i0,int n0,int* l0)
{
	int i=*i0+n0;
	int l=*l0-n0;
	if (text[i]=='<') return;
	while (l) {
		char c=text[i];
		if (c=='<') { 
			int n=MeasureTag(text+i);
			i+=n;

		} else if (c=='&') {
			int x;
			int n=MeasureChar(text+i,&x);
			i+=n;
			l-=x;
		} else 
		{
			i++;
			l--;
		}
	}
// Should also cope with ignorable characters and entities
	*l0=i-*i0;
}






char *GetRef(sara_ulong code, sara_long pos, WORDNO lw, int nw, int nw0,
             int *ipos, int *len, char *scope, char *sno)
{
   char        path[_MAX_PATH] = "";
   int         i = 0;
   FILE       *fText;
   sara_long   l1 = 0;
   sara_long   l2 = 0;
   sara_long   ll;
   sara_long   ls = 0;
   char        id[MAXNAME + 1] = "";
   BOOL        bEll1 = FALSE;
   BOOL        bEll2 = FALSE;
   int         it = 0;
   sara_long   lk;
   int         i0;

   path[0] = '\0';
   GetDocPath(code, id, path);
   BoundElement(scope, lw, code, &l1, &l2, &it, &ls, pos, nw);

   if (l1 > pos)
      l1 = pos;



   if (l2 - l1 > max_sent) {
      sara_long   ls1;
      sara_long   ls2;

      if (pos < l1 + max_sent / 2 - 4) {
         l2 = l1 + max_sent - 4;
         lk = -1;
         bEll2 = TRUE;
      } else {
         l1 = pos - max_sent / 2;
         bEll1 = TRUE;
         if (l2 - l1 > max_sent) {
            l2 = l1 + max_sent - 4;
            lk = -1;
            bEll2 = TRUE;
         }
      }
      ls2 = LastSGMLp(l2, code, defscope, ID_TAGO);
      ls1 = NextSGMLp(l1, code, defscope, ID_TAGO);
      if (ls1 < pos)
         l1 = ls1;
      if (ls2 > pos)
         l2 = ls2;
	  assert(l2-l1<sizeof psent);
   }

   if ((fText = fopen(path, "r")) == NULL) {
#if defined(__unix__)   
      fprintf(stderr, "Can't open %s\n", path);
      perror("fopen()");
#endif      
      return 0;
   }
   fseek(fText, (off_t)l1, SEEK_SET);
   ll = fread(psent, 1, (size_t) (l2 - l1), fText);
   /* Next get the sentence number */
/*   {
      char *s;
      char  psno[200];

      if (l1 <= ls && ls + 10 < l2)
         s = psent + (int) (ls - l1);
      else {
         fseek(fText, (off_t)ls, SEEK_SET);
         fread(psno, 1, 100, fText);
         s = psno;
      }
      if (strstr(s, "<s n=\"") == s)
         sscanf(s,"<s n=\"%[^\"]",sno);
	  else if (strstr(s, "<s n=") == s) 
         sscanf(s,"<s n=%[0-9]",sno);
      else
         strcpy(sno,"?");
   }
*/
   fclose(fText);
   if (!GetAttrib(code,pos,labels[0].element,labels[0].attribute,sno)) strcpy(sno,"?");
   *ipos = (int) (pos - l1);
   *len = nw;
   AdjustLength(psent,ipos,nw0,len);
   if (ll < l2 - l1)
      l2 = l1 + ll; 
/*   lk = l2;
   if (it == LG_ELEMENT)
      lk -= 20;   
   if (lk < pos+nw) lk=pos+nw; 
   if (it == LG_ELEMENT && lk > l1) {
      for (i0 = (int) (lk - l1); i0 < (int) (l2 - l1); i0++)
         if (psent[i0] == '>')
            break;
      l2 = l1 + i0 + 1;
   }           
*/   if (it==LG_EMPTY) {
   		for (i0=0;i0<(int)(l2-l1);i0++)
   			if (psent[i0]=='>')
   				break;
   		l2=l1+i0+1;				
   }				
   psent[l2 - l1] = 0;
   return psent;
}


void TextExtract(sara_ulong text, sara_long l1, sara_long l2, char *sol)
{
   char      id[10];
   char      path[_MAX_PATH] = "";
   FILE     *fText;
   int       j = 0;
   int       i1 = 0;
   int       i2 = 0;

   GetDocPath(text, id, path);

   if ((fText = fopen(path, "r")) == NULL) {
      syslog(LOG_ERR, "Can't open %s", path);
      return;
   }
   fseek(fText, (off_t)l1, SEEK_SET);
   fread(psent, 1, (size_t) (l2 - l1), fText);
   fclose(fText);
   psent[l2 - l1] = 0;

   Trim(psent, &i1, &i2);

   while (psent[j] != '>')
      j++;

   strcpy(sol, psent + j + 1);
}                   

#define MAXPOS 10000000
#define MAXEE 200

void BoundElement(char *el, WORDNO nw, sara_ulong code, sara_long *l1,
                  sara_long *l2, int *it, sara_long *ls, sara_long pos,
                  sara_long len)
{
   int         i;
   sara_long   ltt = -1;
   sara_long   lt;
   int         ik;
   char       *s;
   int         ns = 0;
   char       *w = el;
   char        el0[20];
   char sent[2];
   strcpy(sent,"s");


   el0[0] = '\0';
   if (isdigit(*w)) {
      ns = atoi(w);
      *l1 = LastSGML(nw - ns, code, defscope, ID_TAGO);
      *ls = LastSGML(nw, code, defscope, ID_TAGO);
      *l2 = NextSGML(nw + 1, code, defscope, ID_TAGO);
      *it = LG_SENT;
      return;
   }
   while (s = strtok(w, ",")) {
      w = NULL;
/*
      if (i==nelts)
         continue;
 */
      lt = LastSGMLp(pos, code, s, ID_TAGO);
      if (lt > ltt) {
         strcpy(el0, s);
         ltt = lt;
      }
   }
   if (!el0[0]) {
 /*     strcpy(el0, DOCTYPE); */
/*      ltt = LastSGMLp(pos, code, el0, ID_TAGO); */
		ltt=0;
   }
   for (i = 0; i < nelts; i++)
      if (SGMLComp(eltab[i].name, el0))
         break;
   *it = ik = (i < nelts) ? eltab[i].eltype : LG_SENT;   /* LG_EMPTY was LG_SENT */

   switch (ik) {
    case LG_SENT:  
    	*ls=ltt;
      *l1 = ltt;
//    panic:
      *l2 = NextSGMLp(pos + len, code, sent, ID_TAGO);
      return;

    case LG_EMPTY:
      *l1 = ltt;
/*      *l2 = NextSGML(nw + 1, code, DOCTYPE, ID_TAGE); */
	*l2=MAXEE+ltt;
      *ls = LastSGML(nw, code, defscope, ID_TAGO);
      return;

    case LG_ELEMENT:
      *l1 = ltt;
      *l2 = NextSGMLp(*l1, code, el0, ID_TAGE);
      *ls = LastSGMLp(pos, code, defscope, ID_TAGO);
      if (*l2 < pos + len) {
      	*l2 = NextSGMLp(pos + len, code, defscope, ID_TAGO);
		*it=LG_SENT;
		}		
      break;
   }
}

/*
void BoundElement2(char *el, WORDNO nw, sara_ulong code, sara_long *l1,
                   sara_long *l2, int *it, sara_long *ls, sara_long pos,
                   sara_long len)
{
   int   i;
   int   ik;

   for (i = 0; i < nelts; i++)
      if (!strcmp(eltab[i].name, el))
         break;

   *it = ik = (i < nelts) ? eltab[i].eltype : LG_SENT;

   switch (ik) {
    case LG_SENT:
    panic:
      *ls = *l1 = LastSGMLp(pos, code, "s", ID_TAGO);
      *l2 = NextSGMLp(pos + len, code, "s", ID_TAGO);
      return;

    case LG_EMPTY:
      *l1 = LastSGML(nw, code, el, ID_TAGO);
      *l2 = NextSGML(nw + 1, code, DOCTYPE, ID_TAGE);
      *ls = LastSGML(nw, code, "s", ID_TAGO);
      return;

    case LG_ELEMENT:
      *l1 = LastSGMLp(pos, code, el, ID_TAGO);
      *l2 = NextSGMLp(*l1, code, el, ID_TAGE) + 20;
      if (*l2 < pos + len)
         goto panic;

      *ls = LastSGMLp(pos, code, "s", ID_TAGO);
      break;
   }
}

*/


#if defined(SERVER)

int readrec(FILE *f, char *a, char *b, char *c)
{
   if (!readst(f, a))
      return 0;

   readst(f, b);
   readst(f, c);
   return 3;
}

BOOL readst(FILE *f, char *n)
{
   char  c;

   while (isspace(c = fgetc(f)))
      ;

   if (c == EOF)
      return FALSE;

   do {
      *(n++) = c;
      c = fgetc(f);
   } while (!isspace(c) && c != EOF)
      ;

   *n = 0;

   return TRUE;
}

#endif /* SERVER */

struct {
   BOOL      bUse;
   char      name[10];
	enum filtype cmd;
}

__far filtab[MAXFILTERN];

void ClearFilters(char *name)
{
   int   i;
   int   j = 0;

   for (i = 0; i < MAXFILTERN; i++)
      if (filtab[i].bUse && strcmp(name, filtab[i].name))
         filtab[j++] = filtab[i];

   for (i = j; i < MAXFILTERN; i++)
      filtab[i].bUse = FALSE;

   bSLen = FALSE;
}


void AddFilter(char *name, struct pcmd *p)
{
   int   i;

   for (i = 0; i < MAXFILTERN; i++)
      if (!filtab[i].bUse)
         break;

   if (i == MAXFILTERN) {
      SetString("NO MAXFILTERS");
      return;
   }

   filtab[i].bUse = TRUE;
   strcpy(filtab[i].name, name);
   filtab[i].cmd = p->i;

   return;
}

BOOL WTag0(const char* s)
{
	char b[MAXTAG]; char* ss=b;
	if (*s++ != '<') return FALSE;
	while (*s && *s != ' ') *ss++=*s++;
	*ss=0;
	return IsWTag(b);
}

void Reformat(char *name, char *s, int *pos, int *len)
{
   int   i;

   for (i = 0; i < MAXFILTERN; i++) {
      if (!filtab[i].bUse)
         continue;

      if (strcmp(filtab[i].name, name))
         continue;

      ApplyFilter(filtab[i].cmd, s, pos, len);
   }

   if (*(s + *pos) == '<') {
	   if (WTag0(s+*pos)) {
		   int k=2;
		   while (*(s+*pos+k)) {
			   if (WTag0(s+*pos+k)) {*len=k; return;}
				k++;
		   }
	   }

      (*pos)++;             
      if (*(s+*pos) =='/') (*pos)++;
   }   
}


void ApplyFilter(enum filtype i, char *s, int *pos, int *len)
{
   switch (i) {
    case FIL_SGMLX:
      StripAllSGML(s, pos, len);
      break;

    case FIL_NOPOSX:
      if (!bNoPos) StripAllPOS(s, pos, len);
	  else nPOSx=0;
      break;

    case FIL_NORMSPACE:
      NormSpace(s, pos, len);
      break;

/*    case FIL_LEFTL:
      LeftL(s, pc->arg[0].num, pos, len);
      break;

    case FIL_ADDSLEN:
      bSLen = pc->arg[0].num;
      break;
*/
    case FIL_CMAP:
      TranslateChars(s, pos, len);
      break;
/*
    case FIL_ADDPOS:
      bPOS = pc->arg[0].num;
      break;

    case FIL_ADDTEXT:
      bText = pc->arg[0].num;
      break;
*/
    case FIL_NORMCR0:
      NormCR0(s, pos, len);
      break;

    case FIL_ADJSGML:
      break;

    case FIL_ADJPOS:
      {
         int   i = *pos + *len;
     	  if (bNoPos) break;

         if (s[i] == '&') {
            while (s[i] != ';' && s[i] != 0)
               i++;

            if (s[i] == ';')
               i++;
         }
         *len = i - *pos;
         break;
      }

    case FIL_WTOPOS:
      if (!bNoPos) WToPos(s, pos, len);
      break;
   }
}


void StripAllSGML(char *s, int *pos, int *len)
{
   DelSGML(s, pos, len);
}



void DelSGML(char *line, int *pos, int *len)
{
   char *s = line;
   char *s1 = line;
   int   k;
   int   pose = *pos + *len;

   BOOL  bSG = FALSE;
   BOOL  bCh = FALSE;
   int   newpos = -1;
   int   newpose = -1;

   while (*s != 0) {
      if (k = IsSGML(s)) {
         s += k;
         bSG = TRUE;
      } else if (*s == 10) {
         if (bSG && !bCh)
            s++;
         else
            *(s1++) = *(s++);

         bCh = FALSE;
         bSG = FALSE;
      } else if (*s == 13) {
         if (bSG & !bCh)
            s++;
         else
            *(s1++) = *(s++);
      } else {
         bCh = TRUE;
         *(s1++) = *(s++);
      }

      if (newpos < 0 && s >= line + *pos)
         newpos = s1 - line;

      if (newpose < 0 && s >= line + pose)
         newpose = s1 - line;
   }

   *s1 = 0;
   *pos = newpos;
   *len = newpose - newpos;

   return;
}


void WToPos(char *line, int *pos, int *len)
{
   char      b1[MAX_SOL] = "";
   char     *s = line;
   char     *s1 = b1;
   int       pose = *pos + *len;
   int       newpos = -1;
   int       newpose = -1;
   char      pos0[10];
   char     *pos1;
   BOOL      bWP;
   int       i;
   int       x;
   int       k;

   bWP = FALSE;

   while (TRUE) {
      if ((*s == '<' || *s == 0) && bWP) {
         x = strlen(pos0);
         k = 0;

         while (isspace(*(s1 - k - 1)))
            k++;

         s1 -= k;

         for (i = 0; i < k; i++)
            *(s1 + x + i + 2) = *(s1 + i);

         *s1 = '&';

         for (i = 0; i < x; i++)
            *(s1 + i + 1) = *(pos0 + i);

         *(s1 + x + 1) = ';';

         bWP = FALSE;

         s1 += x + 2 + k;
      }

      if (*s == 0)
         break;

      if (*s == '<') {
         int   i = 0;

         while (s[i]) {
            if (s[i] == '>')
               break;
            i++;
         }

         if (s[i] == 0)
            break;
      }
      if (   *s == '<'
          && (*(s + 1) == 'w' || *(s + 1) == 'c') && *(s + 2) == ' ') {
         s += 3;
         pos1 = pos0;

         while (*s != '>')
            *(pos1++) = *(s++);

         *pos1 = 0;
         s++;
         bWP = TRUE;
      } else
         *(s1++) = *(s++);

      if (newpos < 0 && s >= line + *pos)
         newpos = s1 - b1;

      if (newpose < 0 && s >= line + pose)
         newpose = s1 - b1;
   }

   *s1 = 0;
   strcpy(line, b1);

   *pos = newpos;
   *len = newpose - newpos;

   return;
}


int IsSGML(char *s)
{
   BOOL      b;
   int       k = 0;
   char      c;
   char     *s1;

   if (*s != '<')
      return 0;

   if (*(s + 1) == '/')
      k = 1;

   c = *(s + k + 1);

   if (!isalpha(c))
      return 0;

   s1 = s;

   while (*s1 != '>' && (b = (*s1 != 0)))
      s1++;

   return b ? s1 - s + 1 : s1 - s;
}


void StripAllPOS(char *s, int *pos, int *len)
{
   nPOSx = DelPOS(s, pos, len);
}


int DelPOS(char *line, int *pos, int *len)
{
   char     *s = line;
   char     *s1 = line;
   int       i;
   int       id = 0;
   int       pose = *pos + *len;

   while (*s != 0) {
      s += IsPOS(s, &i);

      if (*s == 0)
         break;

      if (s == line + *pos)
         *pos = s1 - line;

      if (s == line + pose)
         pose = s1 - line;

      *(s1++) = *(s++);

      if (s == line + pose) {
         pose = s1 - line;
         if (!IsPOS(s, &id))
            id = 0;
      }
   }

   *s1 = 0;
   *len = pose - *pos;

   return id;
}


int IsPOS(char *s, int *id)
{
   int   i;
   int   k;

   if (*s != '&')
      return 0;

   for (i = npos- 1; i >= 0; i--)
      if (strtest(postab[i].name, s + 1))
         break;

   if (i < 0)
      return 0;

   k = strlen(postab[i].name) + 1;

   if (*(s + k) == ';')
      k++;

   *id = i;

   return k;
}


BOOL strtest(char *test, char *string)
{
   char *t = test;
   
   while (*t != 0)
      if (*(t++) != *(string++))
         return FALSE;
	if (isalnum(*string)) return FALSE;
   return TRUE;
}


void NormSpace(char *s, int *pos, int *len)
{
   DelNL(s, pos);
   ConsolSpaces(s, pos);
}


void ConsolSpaces(char *line, int *pos)
{
   char     *s = line;
   char     *s1 = line;
   BOOL      bSpace = FALSE;

   while (*s != 0) {

      if (isascii(*s) && isspace(*s)) {
         if (!bSpace)
            *(s1++) = ' ';

         bSpace = TRUE;
      } else {
         *(s1++) = *s;
         bSpace = FALSE;
      }

      if (s == line + *pos)
         *pos = s1 - line - 1;

      s++;
   }

   if (s == line + *pos)
      *pos = s1 - line;

   *s1 = 0;

   return;
}


void DelNL(char *line, int *pos)
{
   char     *s = line;
   char     *s1 = line;

   while (*s != 0) {
      if (s == line + *pos)
         *pos = s1 - line;

      if (*s == 13 && *(s + 1) == 10) {
		if (oldcr) *(s1++)=' ';
         s++;
      } else if (*s=='\n') 
		{if (oldcr) *(s1++)=' ';}
      else   
         {*s1 = *s;  s1++;}    

      s++;
/*      s1++; */
   }

   if (s == line + *pos)
      *pos = s1 - line;
   *s1 = 0;

   return;
}


void LeftL(char *s, int n, int *pos, int *len)
{
   int   i;

   if (*pos > n) {
      int   k = *pos - n;

      for (i = 0; i < MAX_SOL - k; i++)
         s[i] = s[i + k];

      for (i = MAX_SOL - k; i < MAX_SOL; i++)
         s[i] = ' ';
   } else {
      int   k = n - *pos;

      for (i = MAX_SOL - 1; i >= k; i--)
         s[i] = s[i - k];

      for (i = k - 1; i >= 0; i--)
         s[i] = ' ';
   }

   *pos = n;
}


int chtab[MAXCENTS];

void ClearChars(void)
{
   int   i;

   for (i = 0; i < MAXCENTS; i++)
      chtab[i] = -1;
}


void AddChar(char *n, int k)
{
   int   i;

   for (i = 0; i < nctab; i++)
      if (!strcmp(n, ctab[i].cname))
         break;

   if (i == nctab)
      return;

   chtab[i] = k;
}


void TranslateChars(char *line, int *pos, int *len)
{
   char *s0 = malloc(strlen(line)+1); char* s=s0;
   char *s1 = line;
   int   k;
   int  c;
   int   pose = *pos + *len;

   strcpy(s,line);

   while (*s != 0) {
      if (s == s0 + *pos)
         *pos = s1 - line;

      if (s == s0 + pose)
         pose = s1 - line;

      if (k = IsSpec(s, &s1))
         s += k;

      else if ((k = IsChar(s, &c)) && (c != -1)) {
         s += k;
		 if (c) {
			*(s1++) = '\x15';
			*(s1++) = '0';
			*(s1++) = '0';
			*(s1++) = (c/16 > 9) ? (c/16-9+64):(c/16+'0');
			*(s1++) = (c%16 > 9) ? (c%16-9+64):(c%16+'0');
		 }
      } else
         *(s1++) = *(s++);
   }
      if (s == s0 + *pos)
         *pos = s1 - line;

      if (s == s0 + pose)
         pose = s1 - line;

   *s1 = 0;
   *len = pose - *pos;
	free(s0);
   return;
}


int IsSpec(char *s, char **s1)
{
   int   i;
   int   k;

   if (*s != '&')
      return 0;

   for (i = 0; i < nspec; i++)
      if (strtest(specs[i].name, s + 1))
         break;

   if (i == nspec)
      return 0;

   k = strlen(specs[i].name) + 1;
   if (*(s + k) == ';')
      k++;
   TranslateSpec(i, s1);

   return k;
}


void TranslateSpec(int i, char **s)
{
	char* ss=specs[i].disp;
	while (*ss)
		*((*s)++)=*(ss++);
}


int IsChar(char *s, int *c)
{
   int   i;
//   int   j;
   int   k;

   if (*s != '&')
      return 0;

   for (i = nctab - 1; i >= 0; i--)
      if (strtest(ctab[i].cname, s + 1))
         break;

   if (i < 0)
      return 0;

   k = strlen(ctab[i].cname) + 1;
   if (*(s + k) == ';')
      k++;
/*   *c = 0;
   j = ctab[i].val;
   if (j != -1)
      *c = j;
*/
	*c=ctab[i].val;

   return k;
}


void NormCR0(char *s, int *p, int *l)
{
   char     *s1 = s;
   char      sx[MAX_SOL] = "";
   char     *s2 = sx;
   int       p1 = *p + *l;
   int       np0,
             np1;
   int       i;

   while (*s1 != 0 && s2 - sx < MAX_SOL - 10) {
      if (s1 == s + *p)
         np0 = s2 - sx;

      if (s1 == s + p1)
         np1 = s2 - sx;

      if (*s1 == 13 && *(s1 + 1) == 10) {
         *(s2++) = *(s1++);
         *(s2++) = *(s1++);
      } else if (*s1 == 10) {
         *(s2++) = 13;
         *(s2++) = *(s1++);
      } else
         *(s2++) = *(s1++);
   }

   *p = np0;
   *s2 = 0;
   *l = np1 - *p;

   for (i = 0; i < MAX_SOL - 1; i++) {
      if (sx[i] == 0)
         break;
      s[i] = sx[i];
   }

   s[i++] = 0;
   return;
}


void corperr(char *s, ...)
{
   return;
}


sara_long MatchTags(sara_long pos, int ifx, char* nm)
{
	int k=1;
	sara_long ls, le;
	while (k>0) {
		le=NextSGMLp(pos+1, ifx, nm, ID_TAGE);
		ls=NextSGMLp(pos+1, ifx, nm, ID_TAGO);
		if (ls < le) {
			k++;
			pos=ls;
		} else if (le < ls) {
		k--;
		pos=le;
		} else break;
	}
	return pos;
}





char *GetHeadElt(char *text, sara_long pos, int d, sara_long *newpos,
                 int *newd, BOOL * bj, sara_long *jump)
{
   char      path[_MAX_PATH] = "";
   FILE     *f;
   int       Ty = 0;
   int       k = 0;
   char      nm[10];
   int       ifx;

   *bj = FALSE;
   *jump = 0;

   ifx = FindDoc(text);
   GetDocPath(ifx, text, path);
   
   if ((f = fopen(path, "r")) == NULL) {
      syslog(LOG_ERR, "Can't open %s", path);
      return NULL;
   }


/*   if (pos == 0)
      pos = NextSGMLp(pos, ifx, "txt0", ID_TAGO);
*/
/*
 * FindHeader(f);
 */

 x:
   fseek(f, (off_t)pos, SEEK_SET);
	
   k = GetElt(f, psent, &Ty, nm);
//	assert(k>0);
   *newpos = pos + k;

   if (Ty == -1) {
      pos = *newpos;
      goto x;
   } else if (Ty == ID_WORD)
      *newd = d;
   else if (Ty == ID_TAGO) {
      *newd = d;                /* was d+1 */
      *bj = TRUE;
//      *jump = NextSGMLp(pos, ifx, nm, ID_TAGE);
	  *jump=MatchTags(pos,ifx,nm);
      ETag(f, jump);
   } else {
      pos = *newpos;
      d--;
      goto x;
   }

   fclose(f);

   return psent;
}


int GetElt(FILE *f, char *buff, int *ty, char *nm)
{
   char     *b = buff;
   int       k;

   k = fread(buff, 1, MAXSENTENCE, f);
   buff[k] = 0;
   while (*b && isspace(*b))
         b++;
   if (*b==0) {*ty = ID_WORD; return 0;}

   if (*b == '<' && !wtag(b)) {
      b++;
      if (*b == '/') {
         b++;
         GetName(&b, nm);
         if (Ignore(nm))
            *ty = -1;
         else
            *ty = ID_TAGE;
      } else {
         GetName(&b, nm);
         if (Ignore(nm))
            *ty = -1;
         else if (Empty(nm))
            *ty = ID_WORD;
         else
            *ty = ID_TAGO;
      }

      while (*b != '>')
         b++;
      b++;
      while (*b && isspace(*b))
         b++;

      *b = 0;
      return b - buff;
   } else {
      while (*b != 0 && (*b != '<' || wtag(b)))
         b++;

      *b = 0;
      *ty = ID_WORD;

      return b - buff;
   }
}

BOOL namechar(char c)
{
	if (isalnum(c)) return TRUE;
	if (c=='-' || c=='.') return TRUE;
	return FALSE;
}

BOOL wtag(char *s)
{
	int i, j; char t0[MAXNAME]; char* t=t0;
	if (s[0] != '<') return FALSE;
	j=1;
	if (s[j]=='/') j++;
	while (isspace(s[j])) j++;
	while (namechar(s[j])) *t++=s[j++];
	*t=0;
	if (!bNoPos && IsWTag(t0)) return TRUE;
	for (i=0;i<nelts;i++) 
		if (SGMLComp(eltab[i].name,t0)) break;
	if (i==nelts) return FALSE;
	return eltab[i].bInternal;
}


BOOL Empty(char *nm)
{
   int   i;

   for (i = 0; i < nelts; i++)
      if (SGMLComp(eltab[i].name, nm))
         break;

   return eltab[i].eltype == LG_EMPTY || eltab[i].eltype == LG_SENT;
}


sara_long FindHeader(FILE *f)
{
   int   bx = 0;
   char  b[MAXSENTENCE];
   char *b1;

 y:
   fread(b, 1, MAXSENTENCE, f);

   b1 = strstr(b, "<" DOCTYPE);
   if (b1 == NULL) {
      bx += MAXSENTENCE / 2;
      fseek(f, (off_t)bx, SEEK_SET);
      goto y;
   }
   return bx + b1 - b;
}


BOOL Ignore(char *s)
{
   return FALSE;
}


void ETag(FILE *f, sara_long *l)
{
   int       i;
   char      bu[100];

   fseek(f, *(off_t *)l, SEEK_SET);
   fread(bu, sizeof(bu), 1, f);

   for (i = 0; i < sizeof(bu); i++)
      if (bu[i] == '>')
         break;
	i++;
	while (isspace(bu[i])) i++;
   *l += i;
}


void Trim(char *s, int *j, int *j1)
{
   int       i = strlen(s) - 1;
   int       k = 0;

   if (i<0) return;
   while (isspace(s[i]) && i > 0)
      i--;

   if (i > 0)
      i++;

   s[i] = 0;

//   TranslateChars(s, j, j1);
//   NormSpace(s, j, j1);

   while (s[k] == ' ') {
      (*j)--;
      k++;
   }
}


void GetDictEntry(FILE *f, sara_long k, char *wd, int maxwd, sara_long *freq, sara_long *forms)
{
   struct di rec[2];
   off_t     kk = (off_t)k * sizeof(rec[0]) + sizeof(struct dhead);

   fseek(f, kk, SEEK_SET);
   fread(rec, sizeof(rec[0]), 2, f);
   fix_di(rec, 2);

   *freq = rec[0].freq;
   strcpy(wd,GetWordList(rec[0].w));
   *forms=rec[1].str-rec[0].str;

//   fseek(f, (off_t)rec.w, SEEK_SET);
//   fread(wd, 1, maxwd, f);
}


char        db[1000];
int         sb = 1000;
int         kb;
int         nb;
sara_long   lb;
int xbx, xby;

int GetFirstDictEntry(char *s)
{
   struct di rec;
   struct di recl;
   sara_long x1;
   sara_long x2;
   off_t     kk = sizeof(struct dhead);

   GetLims(fDict, 0, ndix - 1, s, &x1, &x2);

   fseek(fDict, (off_t)(kk + x1 * sizeof(rec)), SEEK_SET);
   fread(&rec, sizeof(rec), 1, fDict);
   fix_di(&rec, 1);
   fseek(fDict, (off_t)(kk + x2 * sizeof(rec)), SEEK_SET);
   fread(&recl, sizeof(recl), 1, fDict);
   fix_di(&recl, 1);
   fseek(fDict, (off_t)rec.w, SEEK_SET);

   kb = nb = sb;
   lb = (recl.w - rec.w) + kb;
	xbx=x1; xby=x2;
   return (int) x1;
}



char *GetNextDictEntry()
{
	static char w[MAXWORD]; sara_long f, f0;
	if (xbx > xby) return NULL;
	GetDictEntry(fDict, xbx, w, MAXWORD, &f, &f0);
	xbx++;
	return w;
}

/*

char *GetNextDictEntry()
{
   int   i;
   int   j;

   if (kb > lb)
      return NULL;

   for (i = kb; i < nb; i++)
      if (db[i] == 0)
         break;

   if (i < nb) {
      int   x = kb;

      kb = i + 1;

      return db + x;
   }

   if (nb < sb)
      return NULL;

   for (j = kb; j < nb; j++)
      db[j - kb] = db[j];

   nb = fread(db + nb - kb, 1, kb, fDict) + j - kb;
   lb -= kb;
   kb = 0;

   return GetNextDictEntry();
}
*/


BOOL GetLims(FILE *f, sara_long i1, sara_long i2, const char *patt,
             sara_long *x1, sara_long *x2)
{
   sara_long   mid;
   char        s[100] = "";
   int         c;
   sara_long   fr, fo;

   if (i1 > i2)
      return FALSE;

   mid = (i1 + i2) / 2;

   GetDictEntry(f, mid, s, sizeof(s), &fr, &fo);

#if defined(__unix__)
   c = dcmp((signed char *)s, (signed char *)patt);
#else /* ! __unix__ */
   c = dcmp((unsigned char *)s, (unsigned char *)patt);
#endif /* __unix__ */
   if (c < 0)
      return GetLims(f, mid + 1, i2, patt, x1, x2);
   else if (c == 0) {
      if (i1 == i2) {
         if (x1)
            *x1 = i1;

         if (x2)
            *x2 = i1;

         return TRUE;
      } else if (i1 == mid) {
         if (x1)
            *x1 = i1;

         if (x2)
            *x2 = GetLims(f, i2, i2, patt, NULL, NULL) ? i2 : i1;

         return TRUE;
      } else {
         if (x1)
            if (!GetLims(f, i1, mid, patt, x1, NULL))
				*x1=mid;

         if (x2)
            if (!GetLims(f, mid, i2, patt, NULL, x2))
				*x2=mid;

         return TRUE;
      }
   } else
      return GetLims(f, i1, mid - 1, patt, x1, x2);
}


int ccmp(unsigned char c1,unsigned char c2)
{
//	unsigned int a1=c1;
//	unsigned int a2=c2;
//	if (a1 != a2) return a1-a2;
	return c1-c2;
}
// this seems to be what DOS sort does

#if defined(__unix__)
int dcmp(signed char *s1, signed char *s2)
#else /* ! __unix__ */
int dcmp(unsigned char *s1, unsigned char *s2)
#endif /* __unix__ */
{
	int k;
	while (*s2 != 0) {
	   k=uccmp(*s1,*s2);
      if (k<0)
         return -1;
      else if (k>0)
         return 1;
      else {
         s1++;
         s2++;
	}}

   return 0;
}


BOOL NewQName(char *name)
{
   int   i;

   for (i = 0; i < NFILES; i++)
      if (qtable[i].bUse && !strcmp(qtable[i].name, name))
         return TRUE;

   for (i = 0; i < NFILES; i++)
      if (!qtable[i].bUse)
         break;

   if (i == NFILES)
      return FALSE;

   strcpy(qtable[i].name, name);
   sprintf(qtable[i].path, "%s%s.sqy", tmppath, name);

   qtable[i].bUse = TRUE;
   qtable[i].bSave = FALSE;

   return TRUE;
}


char *GetQPath(char *name)
{
   int   i;

   for (i = 0; i < NFILES; i++)
      if (qtable[i].bUse && !strcmp(qtable[i].name, name))
         break;

   if (i == NFILES)
      return NULL;

   return qtable[i].path;
}


void FreeQPath(char *name)
{
   int   i;

   for (i = 0; i < NFILES; i++)
      if (qtable[i].bUse && !strcmp(qtable[i].name, name))
         break;

   if (i == NFILES)
      return;
   qtable[i].bUse = FALSE;
   remove(qtable[i].path);
}


char      tpath[_MAX_PATH];

char *GetQTPath(char *name)
{
   strcpy(tpath, GetQPath(name));
   tpath[strlen(tpath) - 1] = 't';
   return tpath;
}


int OpenQPath(char *name)
{
   int   i;
   int   k = InUse(name);

   if (k == 0)
      return 0;

   for (i = 0; i < NFILES; i++)
      if (!qtable[i].bUse)
         break;

   if (i == NFILES)
      return 0;

   strcpy(qtable[i].name, name);
   sprintf(qtable[i].path, "%s%s.sqy", tmppath, name);
   qtable[i].bUse = TRUE;
   qtable[i].bSave = TRUE;

   return k;
}


BOOL Exists(char *name)
{
   char  path[_MAX_PATH] = "";
   FILE *f;

   sprintf(path, "%s%s.sqy", tmppath, name);

   f = fopen(path, "r");
   if (f != NULL)
      fclose(f);

   return (f != NULL);
}


void Dummy(char *name)
{
   char  path[_MAX_PATH];
   FILE *f;

   sprintf(path, "%s%s.sqy", tmppath, name);

   f = fopen(path, "w");
   fclose(f);
}


int InUse(char *name)
{
   char        path[_MAX_PATH] = "";
   FILE       *f;
   sara_long   l;

   sprintf(path, "%s%s.sqy", tmppath, name);

   f = fopen(path, "r");

   if (f == NULL)
      return 0;

   fseek(f, (off_t)0, SEEK_END);
   l = ftell(f);
   fclose(f);

   return (int) l / sizeof(struct locent1);
}


void CollTable(char *name, int kwin)
{
   char        path[_MAX_PATH] = "";
   FILE       *f;
   sara_long   i;
   sara_long   l;
   char        wd[100];
   sara_long   freq, fo;

   GetCName(name, path);

   f = fopen(path, "w");

   for (i = 0; i < ndix; i++) {
	   int n;
      GetDictEntry(fDict, i, wd, sizeof(wd), &freq, &fo);
      l = colloc_score(name, wd, kwin,kwin,&n);
      fprintf(f, "%s\t%ld\n", wd, l);
   }

   fclose(f);
}


void GetCName(char *name, char *path)
{
   sprintf(path, "%s%s.col", tmppath, name);
}


char      text[MAXTEXTS][MAX_CODE+1];

int FindDoc(char *name)
{
   int   i;

/*
   if (!strcmp(name,"000"))
      return 0;
 */
   for (i = 0; i < ntexts; i++)
      if (!_stricmp(text[i], name))
         break;
   return i;
}


void GetPIndex(void)
{
   char  idxpath[_MAX_PATH];
   int   i;

#if !defined(__unix__)
   _fmode = _O_BINARY;
#endif /* ! __unix__ */

   strcpy(idxpath, etcpath);
   strcat(idxpath, PINDEXNAME);
   fPack = fopen(idxpath, "r");
	if (fPack==NULL) return;
   fread(&idxhead, sizeof(idxhead), 1, fPack);
   fix_pihead(&idxhead, 1);
   ntexts = idxhead.n;

   for (i = 0; i < ntexts; i++) {
      int         j;
      sara_long   x;
      struct pii  ie;
      char        path[_MAX_PATH];

      fread(&ie, sizeof(ie), 1, fPack);
      fix_pii(&ie, 1);
      x = ftell(fPack);
      fseek(fPack, (off_t)ie.path, SEEK_SET);
      fread(path, _MAX_PATH, 1, fPack);
      j = strlen(path);
      if (i == 0)
         strcpy(text[i], "000");
      else {
		  int k;
		  for (k=j-1;k>=0;k--)
			  if (path[k]=='/'||path[k]=='\\') break;
         strcpy(text[i], path+k+1);
	  }
      fseek(fPack, (off_t)x, SEEK_SET);
   }
}

void GetBIndex()
{
	char idxpath[_MAX_PATH]; int k;
	bNoBib=FALSE;
	k=strlen(bibname);
	if (k>0 && bibname[k-1]==SEPCHAR) {
		strcpy(idxpath,bibname);
		strcat(idxpath,BIBNAME);
	} else {
		strcpy(idxpath, etcpath);
		strcat(idxpath, bibname);
	}
   if ((fBib = fopen(idxpath, "r")) == NULL) {   
   		bNoBib=TRUE;
		return;
   }            
   

   strcpy(idxpath, etcpath);
   strcat(idxpath, PBIBNAME);
   if ((fBibIndex = fopen(idxpath, "r")) == NULL) {
   		bNoBib=TRUE;
   }

   return;
}


void GetDocPath(sara_ulong k, char *name, char *path)
{
   struct pii  ie;
   int         i;

   fseek(fPack, (off_t)(sizeof(idxhead) + k * sizeof(ie)), SEEK_SET);
   fread(&ie, sizeof(ie), 1, fPack);
   fix_pii(&ie, 1);
   fseek(fPack, (off_t)ie.path, SEEK_SET);
	strcpy(path,txtpath);
   i = strlen(path);
   fread(path + i, 1, _MAX_PATH - i, fPack);

   i = strlen(path);
   if (k == 0)
      strcpy(name, "000");
   else {
	   int j;
	   for (j=i-1;j >= 0;j--)
		   if (path[j]=='\\' || path[j]=='/') break;
	  strcpy(name,path+j+1);
   }
}


#if !defined(__unix__)
int getpid(void)
{
   return 0;
}
#endif /* ! __unix__ */


BOOL GetBibHeader(char *code, int *ty, int *nr)
{
   off_t          k;
   struct bibhdr  b;
   int            l = 0;
   int            p = 0;
   int            i = FindDoc(code);
   
   if (bNoBib) return FALSE;
   if (i == 0) {
      *ty = 0;
      *nr = 2;
      return TRUE;
   }
   k = i * sizeof(b);
   fseek(fBibIndex, k, SEEK_SET);
   fread(&b, sizeof(b), 1, fBibIndex);
   fix_bibhdr(&b, 1);
   if (b.type == -1)
      return FALSE;
   *ty = b.type;
   *nr = b.nrows;
   return TRUE;
}

#define bncbib1 "The British National Corpus"
#define bncbib2 "Corpus header"

BOOL GetBibItem(char *code, int n, char *buffer)
{
   off_t             k;
   struct bibhdr     b;
   struct bibentry   bi;
   int               l = 0;
   int               p = 0;
   int               i = FindDoc(code);

	if (bNoBib) return FALSE;
   if (i == 0) {
      if (n == 0)
         strcpy(buffer, bncbib1);
      else
         strcpy(buffer, bncbib2);
      return TRUE;
   }
   k = i * sizeof(b);
   fseek(fBibIndex, k, SEEK_SET);
   fread(&b, sizeof(b), 1, fBibIndex);
   fix_bibhdr(&b, 1);
   if (n >= b.nrows)
      return FALSE;
   fseek(fBibIndex, (off_t)(b.seek + n * sizeof(bi)), SEEK_SET);
   fread(&bi, sizeof(bi), 1, fBibIndex);
   fix_bibentry(&bi, 1);
   fseek(fBib, (off_t)bi.name, SEEK_SET);
   fread(buffer, 1, bi.namelen, fBib);
   buffer[bi.namelen] = 0;
   TranslateChars(buffer, &l, &p);
   NormSpace(buffer, &l, &p);
   return TRUE;
}


void LoadChars(char *file)
{
   char  path[_MAX_PATH];
   FILE *f;
   char  buffer[100];

   strcpy(path, etcpath);
   strcat(path, file);
   strcat(path, ".cs");
   if ((f = fopen(path, "r")) == NULL) {
      syslog(LOG_ERR, "Can't open %s", path);
      return;
   }
   ClearChars();

   while (fgets(buffer, sizeof(buffer), f)) {
      char *s = strtok(buffer, " \n\t\r");
      int   k;

      if (s == NULL)
         continue;
      k = atoi(strtok(NULL, " \n\t\r"));
      AddChar(s, k);
   }
   fclose(f);
}


BOOL DTDelt(int n, char *s)
{
   if (n < npos)
      sprintf(s, "POS %s", postab[n].name);
   else if (n < npos + nelts)
      sprintf(s, "ELT %s %d", eltab[n - npos].name,
              eltab[n - npos].eltype);
   else if (n < npos + nelts + natts)
      sprintf(s, "ATT %s %s %d %s", atttab[n - npos - nelts].name,
              atttab[n - npos - nelts].element,
              atttab[n - npos - nelts].attv,
              atttab[n - npos - nelts].attdesc);
   return n < npos + nelts + natts;
}


void ScratchFile(char *s)
{
   int   ifile = 0;
   FILE *f;

   do {
      sprintf(s, "%st%06d", tmppath, ifile++);
   } while (Exists(s));

   f = fopen(s, "w");
   fclose(f);
}

#if defined(__unix__)
/*
 * Detach from our controlling tty
 */

void detach(void)
{
   pid_t pgrp;

   /*
    * if our parent is init (pid 1) we don't have a controlling
    * terminal
    */
   if (getppid() != 1) {
      switch(fork()) {
       case -1:
         syslog(LOG_ERR, "fork() failed [%d]", errno);
         perror("fork() failed");
         exit(1);
         break;

       case 0:
#if !defined(NO_SETSID)
         if ((pgrp = setsid()) == -1) {
            syslog(LOG_ERR, "setsid() failed [%d]", errno);
            perror("setsid() failed");
            exit(1);
         }
#else
         if ((pgrp = setpgrp(getpid(), 0)) == -1) {
            syslog(LOG_ERR, "setpgrp() failed [%d]", errno);
            perror("setpgrp");
            exit(1);
         }
#endif
         break;

       default:
         exit(0);
      }
   }
}                    
#endif /* __unix__ */       

void DelAll(char*,char*);                        

void CleanUp()
{
	DelAll(tmppath,"*.sqy");
	DelAll(tmppath,"*.wl");
}

void DelAll(char* path,char* ext)
{                     
#ifndef __unix__
#ifndef WIN32
	char p[_MAX_PATH];                 
	struct _find_t fi;
	sprintf(p,"%s%s",path,ext);
	if (_dos_findfirst(p,_A_NORMAL,&fi)) return;
	sprintf(p,"%s%s",path,fi.name);
	remove(p);
	while (_dos_findnext(&fi)==0) {
		sprintf(p,"%s%s",path,fi.name);
		remove(p);
	}
#endif
#endif
}		
	
void LowerCase(char* s)
{
	int k=strlen(s);
	int i, j; BOOL bEntity=FALSE;
	for (i=0;i<k;i++) {
		if (s[i]=='&') bEntity=TRUE;
		if (s[i]==';') bEntity=FALSE;
		if (!bEntity) {
			if (isascii(s[i])) s[i]=tolower(s[i]);
			for (j=0;j<ncase;j++) 
				if (lctab[j].u==(unsigned char) s[i]) s[i]=lctab[i].l;
		}
	}
}

struct _sds {
	char w[MAXWORD];
	int f, f1;
};

struct _sds* thedict;

int FTable(int k0,int k1,int k2,const char* patt)
{
	char path[_MAX_PATH]; struct regexp* re;
	FILE* f; char line[1000]; int t=0; int tt; int k;
	if (k0==-1) k0=100000000;
	if (k1==-1) k1=0;
	if (k2==-1) k2=100000000;
	re = ParseRegExp(patt);
	if (re==NULL) return 0;
	GetFTable(path);
	f=fopen(path,"r");
	if (f==NULL) return 0;
	while (fgets(line,sizeof line,f))
	{
		sscanf(line,"%*[^\t]\t%d",&k);
		if (k > k2) continue; 
		if (k < k1) break;
		t++;
		if (t==k0) break;
	}
	thedict=(struct _sds*) malloc(t*sizeof (struct _sds));
	if (thedict==NULL) {
		fclose(f);
		return 0; 
	}
	rewind(f); tt=0;
	while (fgets(line,sizeof line,f) && tt < t)
	{
		char w[MAXWORD]; int f, f1;
		sscanf(line,"%80[^\t]\t%d %d",w,&f,&f1);
		if (f > k2) continue;
		if (f < k1) break;
		if (!pMatch(w,re)) continue;
		strcpy(thedict[tt].w,w);
		thedict[tt].f=f;
		thedict[tt].f1=f1;
		tt++;
		if (tt==k0) break;
	}
	fclose(f);
	return tt;
}

void FEntry(int k,char* buffer,int* freq,int* forms)
{
	strcpy(buffer,thedict[k].w);
	*freq=thedict[k].f;
	*forms=thedict[k].f1;
}

void FFree()
{
	free(thedict);
}

int QuickFrequency(const char* w)
{
	struct stream* s; int k;
	if (SCRegistered()) 
		return DicFrequency(w);
	s=MakeLemmaStream(w); 
	if (s==NULL) return 0;
	k=XCountStream(s);
	FreeStream(s);
	return k;
}

int DicFrequency(const char* w)
{
    sara_long   t1, x1;
    sara_long   t2, x2;
	QADLims(w,&t1,&t2);
    if (GetLims(fDict, t1, t2, w, &x1, &x2)) {
		struct di rec;
		off_t     kk = (off_t)x1 * sizeof(rec) + sizeof(struct dhead);
		fseek(fDict, kk, SEEK_SET);
		fread(&rec, sizeof(rec), 1, fDict);
		fix_di(&rec, 1);
		return rec.freq;
    } else
		return 0;
}

int DicSB(const char* w)
{
    sara_long   x1;
    sara_long   x2;
    if (GetLims(fDict, 0, ndix - 1, w, &x1, &x2)) {
		struct di rec;
		off_t     kk = (off_t)x1 * sizeof(rec) + sizeof(struct dhead);
		fseek(fDict, kk, SEEK_SET);
		fread(&rec, sizeof(rec), 1, fDict);
		fix_di(&rec, 1);
		return rec.str;
    } else
		return 0;
}

void DicAdd(char* w,int k)
{
}

void ReadDictionary()
{
	int i;
	fseek(fDict,sizeof(struct dhead),SEEK_SET);
	for (i=0;i<ndix;i++) {
		struct di rec; int w0; char wd[MAXWORD];
		fread(&rec,sizeof rec,1,fDict);
		fix_di(&rec,1);
		w0=ftell(fDict);
		fseek(fDict,rec.w,SEEK_SET);
		fread(wd,1,MAXWORD,fDict);
		DicAdd(wd,rec.freq);
		fseek(fDict,w0,SEEK_SET);
	}
}

sara_long      Solve0(char *name, int *nt)
{
   FILE             *f;
   sara_ulong        lc = 0xf000000;
   struct locent1    l;
   char             *n;
   sara_long k=0;
   *nt = 0;

   n = GetQPath(name);
   if ((f = fopen(n, "r")) == NULL) {
      syslog(LOG_ERR, "Can't open %s\n", n);
      return 0;
   }
   while (fread(&l,sizeof l,1,f)) {
       if (lc != l.lcode)
            (*nt)++;
	   lc=l.lcode;
	   k++;
   }
   fclose(f);
   return k;
}
		
BOOL Form(char* headword,int k,char* form,int* freq)
{
    sara_long   x1;
    sara_long   x2;
	if (LemmaGetScheme()==0) return FALSE;
    if (GetLims(fDict, 0, ndix - 1, headword, &x1, &x2)) {
		struct di rec; int k1; struct strb sb;
		off_t     kk = (off_t)x1 * sizeof(rec) + sizeof(struct dhead);
		fseek(fDict, kk, SEEK_SET);
		fread(&rec, sizeof(rec), 1, fDict);
		fix_di(&rec, 1);
		k1=rec.str+k;
		fseek(fStreamBase,k1*sizeof sb,SEEK_SET);
		fread(&sb,sizeof sb,1,fStreamBase);
		fix_strb(&sb,1);
		if (sb.frobno==-1) 
			strcpy(form,GetWordList(sb.word));
		else 
			strcpy(form,frobs[fptab[sb.frobno].frob]);
		if (!bNoPos) {
			strcat(form,"=");
			strcat(form,sb.extra==HDRPOS?"HDR":postab[sb.extra%10000].name);
		}
		if (nlt) {
			char t[10];
			sprintf(t, "%d", sb.extra/10000+1);
			strcat(form,"=");
			strcat(form,t);
		}

		*freq=sb.wc;
		return TRUE;
	} else return FALSE;
}
