#include "corpinc.h"

#define MAXSCA 100
#define ALL_NAME "all"

struct _sctab {
	char name[MAXNAME];
	int nt;
	unsigned texts[MAXTEXTS];
	char desc[1000];
	BOOL bReg;
} sctab[MAXSCA];

int FindSubCorpus(const char*);
void RefreshDict();
BOOL Stats(char*,double*,double*,long*,long*);
BOOL GetDictName(char*,char*);

int nsc=0;
int isc=0;

BOOL NewSubCorpus(const char* name)
{
	int k;
	k=FindSubCorpus(name);
	if (k==0) return FALSE;
	else if (k==-1) {
		if (nsc+1 >= MAXSCA) return FALSE;
		strcpy(sctab[nsc].name,name);
		sctab[nsc].texts[0]=0; // header
		k=nsc++;
	} else k--;
	sctab[k].nt=1;
	sctab[k].bReg=FALSE;
	return TRUE;
}

BOOL SetSubCorpus(const char* name)
{
	int k;
	k=FindSubCorpus(name);
	if (k==-1) return FALSE;
	isc=k;
	RefreshDict();
	return TRUE;
}

BOOL SetSubCorpusND(const char* name)
{
	int k;
	k=FindSubCorpus(name);
	if (k==-1) return FALSE;
	isc=k;
	return TRUE;
}

BOOL AddSubCorpus(const char* name,int n, int* p)
{
	int i, k, t;
	k=FindSubCorpus(name);
	if (k<=0) return FALSE;
	k--;
	t=sctab[k].nt;
	for (i=0;i<n;i++) {
		if (p[i]==0) {
			break;
		}
		sctab[k].texts[t++]=p[i];
	}
	sctab[k].nt=t;
	return TRUE;
}

int FindSubCorpus(const char* name)
{
	int i;
	if (!strcmp(name,ALL_NAME)) return 0;
	for (i=0;i<nsc; i++) {
		if (!strcmp(name,sctab[i].name)) return i+1;
	}
	return -1;
}

BOOL TestSubCorpus(unsigned t)
{
	int k, i;
	if (isc==0) return TRUE;
	k=isc-1;
	for (i=0;i<sctab[k].nt;i++) {
		if (t==sctab[k].texts[i]) return TRUE;
		if (t < sctab[k].texts[i]) return FALSE;
	}
	return FALSE;
}

void SCDel(char* name)
{
	int k, i, j;
	k=FindSubCorpus(name);
	if (k<=0) return;
	k--;
	for (i=k+1;i<nsc;i++) {
		strcpy(sctab[i-1].name,sctab[i].name);
		strcpy(sctab[i-1].desc,sctab[i].desc);
		sctab[i-i].nt=sctab[i].nt;
		for (j=1;j<sctab[i].nt;j++) 
			sctab[i-1].texts[j]=sctab[i].texts[j];
		sctab[i-1].bReg=sctab[i].bReg;
	}
	nsc--;
}


BOOL SCSave(char* path)
{
	FILE* f; int n; int i;
	strcpy(path,tmppath);
	strcat(path,sctab[isc-1].name);
	strcat(path,".sc");
	f=fopen(path,"w");
	if (f==NULL) return FALSE;
	n=sctab[isc-1].nt;
	for (i=0;i<n;i++) {
		fprintf(f,"%d\n",sctab[isc-1].texts[i]);
	}
	fclose(f);
	return TRUE;
}

BOOL SCRead(char* name,char* desc)
{
	FILE* f; int k; int t;
	char path[_MAX_PATH]; char line[100];
	strcpy(path,etcpath);
	strcat(path,name);
	strcat(path,".sc");
	f=fopen(path,"r");
	if (f==NULL) return FALSE;
	if (!NewSubCorpus(name)) return FALSE;
	k=FindSubCorpus(name);
	if (k==-1) return FALSE;
	strcpy(sctab[k-1].name,name);
	strcpy(sctab[k-1].desc,desc);
	t=0;
	while (fgets(line,sizeof line,f)) {
		sctab[k-1].texts[t++]=atoi(line);
	}
	sctab[k-1].nt=t;
	sctab[k-1].bReg=TRUE;
	fclose(f);
	return TRUE;
}

void SCReadAll()
{
	FILE* f;
	char path[_MAX_PATH]; char line[1000]; char name[MAXNAME];
	strcpy(path,etcpath);
	strcat(path,"screg");
	f=fopen(path,"r");
	if (f==NULL) return;
	nsc=0;
	while (fgets(line,sizeof line,f)) {
		int k;
		sscanf(line,"%s %n",name,&k);
		SCRead(name,line+k);
	}
	fclose(f);
}

void SCWriteAll()
{
	FILE* f; int i;
	char path[_MAX_PATH];
	strcpy(path,etcpath);
	strcat(path,"screg");
	f=fopen(path,"w");
	if (f==NULL) return;
	for (i=0;i<nsc;i++) {
		fprintf(f,"%s %s\n",sctab[i].name,sctab[i].desc);
	}
	fclose(f);
}

char scbuffer[1000];

int SCCount()
{
	return nsc+1;
}

const char* SCName(int i)
{
	return sctab[i].name;
}

char* SCList(int n)
{
	double fm=0; double fs=0; long nw, nu;
	BOOL bR=Stats(sctab[n-1].name,&fm,&fs,&nw,&nu);	
	sprintf(scbuffer,"%s %d %d %lf %lf %ld %ld %s",sctab[n-1].name,sctab[n-1].nt,bR,fm,fs,nw,nu,sctab[n-1].desc);
	return scbuffer;
}

char* SCEnum(const char* name,int i0,int ik)
{
	int i;
	int k=FindSubCorpus(name);
	scbuffer[0]=0; 
	if (k < 1) return scbuffer;
	for (i=i0;i<i0+ik;i++) {
		char x[10];
		sprintf(x,"%d ",sctab[k-1].texts[i]);
		strcat(scbuffer,x);
	}
	scbuffer[strlen(scbuffer)-1]=0;
	return scbuffer;
}


void GetFTable(char* path)
{
	GetDictName(path,"sdic");
}


BOOL GetDictName(char* path,char* root)
{
//	if (!strcmp(root,"streambase") && LemmaGetScheme()==0) return FALSE;
	strcpy(path,etcpath);
	strcat(path,root);
	strcat(path,".");
	if (isc==0 || !sctab[isc-1].bReg)
		strcat(path,"all.");
	else {
		strcat(path,sctab[isc-1].name);
		strcat(path,".");
	}
	strcat(path,LemmaName(LemmaGetScheme()));
	return TRUE;
}

BOOL Stats(char* name,double* fm,double* fs,long* nw,long* nu)
{
	char path[_MAX_PATH]; FILE *f; struct dhead dh; *nw=*nu=0;
	strcpy(path,etcpath);
	strcat(path,"dic.");
	strcat(path,name);
	strcat(path,".");
	strcat(path,"null");
	f=fopen(path,"r");
	if (f==NULL) return FALSE;
	if (fread(&dh,sizeof(dh),1,f) != 1) {
		fclose(f);
		return FALSE;
	}
//	if (dh.version > 0) {fclose(f); return FALSE;}
	*fm=0;
	*fs=0;
	*nw=dh.nw;
	*nu=dh.nu;

	fclose(f);
	return TRUE;
}



void RefreshDict()
{
   char  dpath[_MAX_PATH];
	if (fDict) fclose(fDict);
   GetDictName(dpath,"dic");

   if ((fDict = fopen(dpath, "r")) == 0) 
		return;
   fread(&dicthead, sizeof(dicthead), 1, fDict);
   fix_dhead(&dicthead, 1);
   ndix=dicthead.nw;
	if (fStreamBase) fclose(fStreamBase);
   GetDictName(dpath,"streambase");
	
	if ((fStreamBase = fopen(dpath, "r")) == 0) 
			return;
	g_uc0=LemmInfo()[2];
	GetDictName(dpath,"qad");
	QADLoad(dpath);
}                                            

BOOL SCRegistered()
{
	return isc == 0 || sctab[isc-1].bReg;
}

long NUsages()
{
	double m, sd; long nw, nu;
	if (isc==0) 
		return dicthead.nu;
	Stats(sctab[isc-1].name,&m,&sd,&nw,&nu);
	return nu;
}

int SBCount(struct locent *l, int k)
{
	int t=0; int i; int p=0;
	if (isc==0) return k;
	for (i=0;i<k;i++, l++) {
		if (l->lcode < sctab[isc-1].texts[p]) continue;
		while (sctab[isc-1].texts[p] < l->lcode) {
			p++;
			if (p >= sctab[isc-1].nt) return t;
		}
		if (sctab[isc-1].texts[p]==l->lcode) t++;
	}
	return t;
}
	
