#include "corpinc.h"

// Changed MAXSIG to MAX_SIG since MAXSIG is defined by Solaris.
//#define MAX_SIG 50
                           
#define MAXTESTLOG 256                           

extern int nroots;                                              
extern char          rpath[][_MAX_PATH];
extern unsigned int  rlim[];
extern char param_file[_MAX_PATH];

enum test_status {TS_OK,TS_WARNING,TS_ERROR,TS_FATAL};
enum test_class {T_PRM,T_DESC,T_LIMS,T_SIG,T_DIRS,T_PDX,T_TEXTS,T_DICT,T_DOWN,
	T_ACC,T_BIB,T_PBIB,T_TEMP,T_MAX};                 

enum test_status TestPRM();        
enum test_status TestDesc();
enum test_status TestLimits();     
enum test_status TestIndexSig();
enum test_status TestIndexDirsExist();
enum test_status TestFilesPdxExists();   
enum test_status TestTexts();                                                                              
enum test_status TestDicExists();
enum test_status TestDownloads();
enum test_status TestAcc();
enum test_status TestBibExists();                                
enum test_status TestPBibExists();                                            
enum test_status TestCS();
enum test_status TestTemp();

   
void TestLog(char*);
void LogPRM(char*);

static enum test_status fStat;
static FILE* fTestLog; 
static int ne, nw;

static char* test_desc[T_MAX]={
"Test PRM settings",                      
"Test corpus description file",     
"Test data limits",            
"Test index signature",
"Test that all index directories exist",
"Test that the compressed file index exists",
"Test that all texts exist",           
"Test that the dictionary exists",
"Test that the downloadable files exist",
"Test that all accelerators exist",
"Test that the bibliography exists",
"Test that the bibliography index exists",
"Test that the temporary directory is writable"};

static char* err_desc[T_MAX]={
"No PRM file",                     
"Couldn't read description file %s",        
"%d limits exceeded",      
"Index signature does not match (%s, %s)",
"Index directory number %d does not exist",
"File %s does not exist",             
"%d texts missing",                   
"File %s does not exist or is wrong version",                    
"%d files missing",      
"%d files missing",
"File %s does not exist",                       
"File %s does not exist or is wrong size",                 
"Cannot create a file in %s"
};         

static enum test_status (*tester[T_MAX])()={
TestPRM,                     
TestDesc,       
TestLimits,                              
TestIndexSig,
TestIndexDirsExist,
TestFilesPdxExists,     
TestTexts,        
TestDicExists,
TestDownloads,  
TestAcc,
TestBibExists,
TestPBibExists,
TestTemp};               
      
static int skip[T_MAX];  

BOOL bLittle()
{	
	char x[4]; int y;
	x[0]=1; x[1]=2; x[2]=3; x[3]=4;
	y=*((int*) x);    
	return (y & 0xff) == x[0];
}
      
void OpenTestLog()
{                                         
	char m[MAXTESTLOG];  time_t t;
	fTestLog=fopen("testlog.txt","wt");
	if (fTestLog==NULL) fStat=TS_FATAL;
	time(&t);
	sprintf(m,"Sara server test log produced %s\n",ctime(&t));
	TestLog(m);
	sprintf(m,"Processor is %s-endian\n",bLittle()?"little":"big");
	TestLog(m);
#if defined(CONV_ALL)
	sprintf(m,"Sara compiled with CONV=ALL\n");
#elif defined(CONV_INDEX)
	sprintf(m,"Sara compiled with CONV=INDEX\n");
#else // CONV_NONE
	sprintf(m,"Sara compiled with no conversion\n");
#endif
	TestLog(m);
}	                                                    

void CloseTestLog()
{
	char m[MAXTESTLOG];
	sprintf(m,"Tests complete. %d errors and %d warnings.\n",ne,nw);
	TestLog(m);
    if (fclose(fTestLog)==EOF) fStat=TS_FATAL;
}                    

enum test_status RunTest(enum test_class t)
{   
	char m[MAXTESTLOG];                   
	if (skip[t]) {
		sprintf(m,"Test %d skipped because of failure of test %d\n",t+1,skip[t]);
		TestLog(m);                                                                  
		return TS_OK;
	}	
	return tester[t]();
}	                                            

BOOL Tests()
{                       
	int i;              
	for (i=0;i<T_MAX;i++)
		skip[i]=0;
	fStat=TS_OK;                 
	ne=nw=0;
	OpenTestLog();
	if (fStat==TS_FATAL) return FALSE;
	for (i=0;i<T_MAX;i++) {
		enum test_status s;
		s=RunTest(i);
		switch (s) {
		case TS_OK:
			break;
		case TS_WARNING:
			nw++;
			break;
		case TS_ERROR:
			ne++;
			break;
		case TS_FATAL:
			fStat=TS_FATAL;
			break;
		}
		if (fStat==TS_FATAL) {
			TestLog("Tests abandoned\n");
			break;
		}	
	}					
	CloseTestLog();
	return ne==0 && fStat==TS_OK;	
}

char* GetTestMsg(enum test_class,int);
                               
void TestLog(char* s)
{
	if (fprintf(fTestLog,s) < 0) fStat=TS_FATAL;
}	                               

static char mb[MAXTESTLOG];

char* GetTestMsg(enum test_class t,int stage)
{                         
	switch (stage) {
	case 0:
		sprintf(mb,"Test %d: ",t+1);
		strcat(mb,test_desc[(int) t]);                
		break;
	case 1:      
		strcpy(mb,err_desc[(int) t]);
		break;
	}	
	strcat(mb,"\n");
	return mb;
}	
                               
void StartTest(enum test_class t)
{
	TestLog(GetTestMsg(t,0));
}	

enum test_status TestLogSuccess()
{
	return TS_OK;
}                                        

char* ecode[TS_FATAL+1]={"OK","WARNING","ERROR","FATAL"};

enum test_status TestLogFail(enum test_class t,...)
{                              
	va_list args;         
	char* fmt; char buffer[MAXTESTLOG]; 
	fmt=GetTestMsg(t,1);                                                         
	strcpy(buffer,"ERR: ");
	va_start(args,t);  
	vsprintf(buffer+strlen(buffer),fmt,args);    
	TestLog(buffer);
	return TS_ERROR;
}	                    

enum test_status TestLogWarn(enum test_class t,...)
{                              
	va_list args;         
	char* fmt; char buffer[MAXTESTLOG]; 
	fmt=GetTestMsg(t,1);                                                         
	strcpy(buffer,"WARNING: ");
	va_start(args,t);  
	vsprintf(buffer+strlen(buffer),fmt,args);    
	TestLog(buffer);
	return TS_WARNING;
}	                    

/* Tests for T_DIRS */

BOOL TestDirExists(char* dir)
{
	int s; char p[_MAX_PATH];
	getcwd(p,_MAX_PATH);
	s=chdir(dir);
	chdir(p);
	return s != -1;
}
	
enum test_status TestIndexDirsExist()
{      
	int i; 
	StartTest(T_DIRS);
	for (i=0;i<NDirs();i++)
		if (!TestDirExists(IndexDir(i))) 
			break;                            
	if (i==NDirs()) 
		return TestLogSuccess();			
	else 	
		return TestLogFail(T_DIRS,i);
}				                           

/* Tests for T_PDX */
BOOL TestFileExists(char* file)
{
	FILE* f=fopen(file,"r");
	BOOL b=(f != NULL);
	if (b) fclose(f);
	return b;
}	
 
enum test_status TestFilesPdxExists()
{
	char  idxpath[_MAX_PATH];
	StartTest(T_PDX);
   	strcpy(idxpath, etcpath);
   	strcat(idxpath, PINDEXNAME);
	if (TestFileExists(idxpath))
		{
		char m[MAXTESTLOG];
		sprintf(m,"There are %d texts",ntexts);
		LogPRM(m);
		return TestLogSuccess();
		}
	else {
		skip[T_TEXTS]=T_PDX+1;
		return TestLogFail(T_PDX,idxpath);
	}	
}		        

/* Tests for T_BIB */

enum test_status TestBibExists()
{
	char  idxpath[_MAX_PATH];
	StartTest(T_BIB);
   	strcpy(idxpath, etcpath);
   	strcat(idxpath, bibname);
	if (TestFileExists(idxpath))
		return TestLogSuccess();
	else {
		skip[T_PBIB]=T_BIB+1;
		return TestLogWarn(T_BIB,idxpath);
	}	
}		        

/* Tests for T_PBIB */

enum test_status TestPBibExists()
{
	char  idxpath[_MAX_PATH];
	StartTest(T_PBIB);
   	strcpy(idxpath, etcpath);
   	strcat(idxpath, PBIBNAME);
	if (TestFileExists(idxpath)) {
		FILE* f=fopen(idxpath,"r");
		struct bibhdr b;                              
		int t;
		fread(&b,sizeof b,1,f);
		fix_bibhdr(&b,1);
		fclose(f);
		t=(int) (b.seek/sizeof b);
		if (t != ntexts) {
			char m[MAXTESTLOG];
			sprintf(m,"Bib index has %d texts",t);
			LogPRM(m);
			return TestLogWarn(T_PBIB,idxpath);
		}	
		return TestLogSuccess();
	}	
	else
		return TestLogWarn(T_PBIB,idxpath);
}		                         

/* Tests for T_DICT */

enum test_status TestDicExists()
{
	char  idxpath[_MAX_PATH];
	StartTest(T_DICT);
   	strcpy(idxpath, etcpath);
   	strcat(idxpath, "dic.all.null");
	if (TestFileExists(idxpath)) {
		struct dhead dh;  
		char m[MAXTESTLOG];
		FILE* f=fopen(idxpath,"r");
		fread(&dh,1,sizeof dh,f);
		fix_dhead(&dh,1);
		fclose(f);
//		if (dh.version > 0) {
//			sprintf(m,"Dictionary has old-format header");
//			LogPRM(m);
//			return TestLogFail(T_DICT,idxpath);
//		}
		sprintf(m,"Dictionary version stamp is %ld",dh.v2);
		LogPRM(m);
		sprintf(m,"%ld words in dictionary",dh.nw);
		LogPRM(m);   
//		return TestLogSuccess();
	}
	else
		return TestLogFail(T_DICT,idxpath);
	strcpy(idxpath,etcpath);
	strcat(idxpath,"streambase.all.null");
	if (!TestFileExists(idxpath)) return TestLogFail(T_DICT,idxpath);
	strcpy(idxpath,etcpath);
	strcat(idxpath,"sdic.all.null");
	if (!TestFileExists(idxpath)) return TestLogFail(T_DICT,idxpath);
	strcpy(idxpath,etcpath);
	strcat(idxpath,"qad.all.null");
	if (!TestFileExists(idxpath)) return TestLogFail(T_DICT,idxpath);
	return TestLogSuccess();
}		                         

/* Tests for T_DOWN */

int ndl=0;

char* dldir[3]={"pagefmt.txt","linefmt.txt","bibfmt.txt"}; 

enum test_status TestDownloads()
{         
	int i; int n=0;
	StartTest(T_DOWN);
	for (i=0;i<ndl;i++) {
		char path[_MAX_PATH];
		strcpy(path,etcpath);
		strcat(path,dldir[i]);
		if (!TestFileExists(path)) {
			char m[MAXTESTLOG];
			n++;                            
			sprintf(m,"File %s does not exist",path);
			LogPRM(m);
		}
	}
	if (n==0) return TestLogSuccess();
	else return TestLogFail(T_DOWN,n);
}			
			
/* Tests for T_ACC */

enum test_status TestAcc()
{         
	int i; int n=0;
	StartTest(T_ACC);
	for (i=0;i<nacc;i++) {
		char path[_MAX_PATH];
		sprintf(path,"%s%d.acc",etcpath,i);
		if (!TestFileExists(path)) {
			char m[MAXTESTLOG];
			n++;                            
			sprintf(m,"File %s does not exist",path);
			LogPRM(m);
		}
	}
	if (n==0) return TestLogSuccess();
	else return TestLogWarn(T_ACC,n);
}			
			
		
/* Tests for T_TEMP */

BOOL TestFileCreate(char* file)
{
	FILE* f=fopen(file,"w");
	BOOL b=(f != NULL);
	if (b) {
		fclose(f);
		remove(file);
	}	
	return b;
}	                                                     

#define BIGFMTMAX 50

void bigfmt(unsigned long b,char s[BIGFMTMAX])
{                                  
	if (b>10000000) 
		sprintf(s,"%dM",b/1000000);
	else if (b> 10000)
		sprintf(s,"%dK",b/1000);
	else 
		sprintf(s,"%d",b);
}	


void DFSReport(char* dir)
{                                             
	char m[MAXTESTLOG]; char s[BIGFMTMAX];
#ifndef __unix__
#ifdef WIN32
	unsigned long b;
	extern BOOL __stdcall GetDiskFreeSpaceA(char*,unsigned*,unsigned*,unsigned*,unsigned*);
	char droot[6]; unsigned spc, bps, ac,tc;
	if (strlen(dir) > 1 && dir[1]==':') {
		droot[0]=dir[0];
		droot[1]=':';
		droot[2]='\\';
		droot[3]=0;
		GetDiskFreeSpaceA(droot,&spc,&bps,&ac,&tc);
	}   
	else
		GetDiskFreeSpaceA(NULL,&spc,&bps,&ac,&tc);
	b=ac;
	b*=spc;
	b*=bps;
#else
	int d=0;               
	struct _diskfree_t df;   unsigned long b;
	if (strlen(dir) > 1 && dir[1]==':') {
		d=dir[0];
		if (d>='a') d-='a'-'A';
		d-='A'-1;
	}   
	_dos_getdiskfree(d,&df);   
	b=df.avail_clusters;
	b*=df.sectors_per_cluster;
	b*=df.bytes_per_sector;
#endif
	bigfmt(b,s);
	sprintf(m,"%s disk free space",s);
	LogPRM(m);
#endif	
}	
	
enum test_status TestTemp()
{
	char  idxpath[_MAX_PATH];
	StartTest(T_TEMP);
   	strcpy(idxpath, tmppath);
   	strcat(idxpath, "$t.sqy");
	if (TestFileCreate(idxpath))
		{
		DFSReport(tmppath);
		return TestLogSuccess();
		}
	else
		return TestLogFail(T_TEMP,tmppath);
}		                         




/* Enumerate PRM settings */

#define NIND 5             

enum par {PAR_NAME,PAR_CNAME,PAR_ETC,PAR_TMP,PAR_ACC,PAR_TXT,PAR_IDX,PAR_BIB,
	PAR_PORT,PAR_ACL,PAR_TOT,PAR_BUF,PAR_LOG,PAR_C6,PAR_OLDNAMES,PAR_OLDID,PAR_WEB,PAR_MAX};

void LogPRM(char* m)
{                      
	char mm[MAXTESTLOG]; 
	int j;
	for (j=0;j<NIND;j++) mm[j]=' ';
	mm[NIND]=0;
	strcat(mm,m);
	strcat(mm,"\n");
	TestLog(mm);
}		
 
void ReportPRM(enum par p)
{      
	int i;                              
	char m[MAXTESTLOG];
	switch (p) {   
	case PAR_NAME:
		sprintf(m,"prm file is %s",param_file);
		LogPRM(m);
		break;
	case PAR_ETC:
		sprintf(m,"etc path is %s",etcpath);
		LogPRM(m);
		break;
	case PAR_TMP:
		sprintf(m,"tmp path is %s",tmppath);
		LogPRM(m);
		break;
	case PAR_ACC:
		sprintf(m,"acc path is %s",accpath);
		LogPRM(m);
		break;
	case PAR_TXT:
		sprintf(m,"txt path is %s",txtpath);
		LogPRM(m);
		break;                                                                                                
	case PAR_BIB:
		sprintf(m,"bib name is %s",bibname);
		LogPRM(m);
		break;
	case PAR_IDX:
		for (i=0;i<nroots;i++) {
			sprintf(m,"index path %d is %s,%d",i,rpath[i],rlim[i]);
			LogPRM(m);
		}
		break;
	case PAR_PORT:
		sprintf(m,"port is %d",port);
		LogPRM(m);
		break;
	case PAR_ACL:               
		for (i=0;i<nacc;i++) {
			sprintf(m,"accelerator %d is %s,%d,%d",i,acctab[i].word,acctab[i].type,acctab[i].extra);
			LogPRM(m);
		}
		break;
	case PAR_TOT:
		sprintf(m,"timeout is %d",timeout);
		LogPRM(m);
		break;
	case PAR_BUF:
		sprintf(m,"nbuff is %d",nbuff);
		LogPRM(m);
		break;
	case PAR_LOG:
		sprintf(m,"logging is %s",bSysLog?"on":"off");
		LogPRM(m);
		break;                                 
	case PAR_CNAME:
		sprintf(m,"corpus name is %s",corpname);
		LogPRM(m);
		break;	                                                              
	case PAR_C6:
		sprintf(m,"C6 compatibility switch is %s",c6?"on":"off");
		LogPRM(m);
		break;
	case PAR_OLDNAMES:
		sprintf(m,"oldnames compatibility switch is %s",oldnames?"on":"off");
		LogPRM(m);
		break;                   
	case PAR_OLDID:
		sprintf(m,"oldid compatibility switch is %s",oldid?"on":"off");
		LogPRM(m);
		break;                   
	case PAR_WEB:
		sprintf(m,"website is %s",website);
		LogPRM(m);
		break;
		 			
	}
}						
										
enum test_status TestPRM()
{         
	int i;                                
	StartTest(T_PRM);
	if (!bInit) {
		TestLogFail(T_PRM);
		return TS_FATAL;
	}
	for (i=0;i<PAR_MAX;i++) 
		ReportPRM(i);                 
	if (nprmerrs > 0) {
		char m[MAXTESTLOG];
		sprintf(m,"There were %d unrecognised lines in the PRM file",nprmerrs);
		LogPRM(m);
	}	
	return TestLogSuccess();	
}		
		
enum test_status TestDesc()
{   
	char n[_MAX_PATH];
	StartTest(T_DESC);  
	strcpy(n,etcpath);
	strcat(n,corpname);
	strcat(n,".dsc");
	if (TestFileExists(n)) 
		return TestLogSuccess();
	else {
		skip[T_LIMS]=T_DESC+1;
		skip[T_SIG]=T_DESC+1;
		return TestLogFail(T_DESC,n);
	}	
}

enum lim {LIM_ELTS,LIM_ATTS,LIM_ENTS,LIM_CHARS,LIM_POS,LIM_MAX};			

BOOL TestLimit(enum lim l)
{
	char m[MAXTESTLOG];
	switch (l) {
	case LIM_ELTS:
		sprintf(m,"%d elements (max %d)",nelts,MAXELTS);
		LogPRM(m);
		return nelts < MAXELTS;
	case LIM_ATTS:
		sprintf(m,"%d attributes (max %d)",natts,MAXGATTS);
		LogPRM(m);
		return natts < MAXGATTS;
	case LIM_ENTS:
		sprintf(m,"%d entities (max %d)",nspec,MAXSPECS);
		LogPRM(m);
		return nspec < MAXSPECS;
	case LIM_CHARS:
		sprintf(m,"%d character entities (max %d)",nctab,MAXCENTS);
		LogPRM(m);
		return nctab < MAXCENTS;
	case LIM_POS:
		sprintf(m,"%d pos codes (max %d)",npos,MAXENTS);
		LogPRM(m);
		return npos < MAXENTS;
	}
	return TRUE; // unused
}		
	
	
enum test_status TestLimits()                                                        
{                                     
	int i; int n=0;                                                                          
	StartTest(T_LIMS);
	for (i=0;i<LIM_MAX;i++)
		if (!TestLimit(i)) n++;
	if (n==0) return TestLogSuccess();
		else
			return TestLogFail(T_LIMS,n);
}				
		
enum test_status TestTexts()
{   
	sara_ulong i; int n=0;
	char name[MAX_CODE+1]; char path[_MAX_PATH];	
	char m[MAXTESTLOG];
	StartTest(T_TEXTS);	
	for (i=0;i<(sara_ulong) ntexts;i++) {
		GetDocPath(i,name,path);  
		if (!TestFileExists(path)) {
			if (omitted(name)) {
				sprintf(m,"Text %s in file %s omitted",name,path);
				LogPRM(m);
			} else {	
				sprintf(m,"Text %s in file %s not found",name,path);
				LogPRM(m);
				n++;
			}	
		}
	}
	if (n==0) return TestLogSuccess();
	else return TestLogFail(T_TEXTS,n);
}	                                                                                        

			
enum test_status TestIndexSig()
{
	char path[_MAX_PATH]; struct index_sig is;
	FILE *f; BOOL bTest; char sig1[MAX_SIG]; char sig2[MAX_SIG];
	StartTest(T_SIG);
	strcpy(path,etcpath);
	strcat(path,"index.sig");                             
	f=fopen(path,"r");                                    
	bTest=FALSE;
	if (f) { 
		if (fread(&is,1,sizeof is,f) == sizeof is) {
				fix_sig(&is,1);
				bTest=(!strcmp(is.name,corpname) && is.nelts==nelts && is.natts==natts && is.npos==npos && (int) is.ntexts==ntexts); 
	    		sprintf(sig2,"%s-%d-%d-%d-%d",is.name,is.nelts,is.natts,is.npos,is.ntexts);                                                       
	    }
	    else 
	    	strcpy(sig2,"bad signature");
	    fclose(f);
	} else
		strcpy(sig2,"no signature");
	if (bTest) return TestLogSuccess();
	nhash=is.nhash;
	nhashsig=is.nhashsig;
	hashlast=is.hashlast;
	hashshift=is.hashshift;
	maxiloc=is.maxiloc;
	sprintf(sig1,"%s-%d-%d-%d-%d",corpname,nelts,natts,npos,ntexts);                                                       
	return TestLogFail(T_SIG,sig1,sig2);
}	
		    	 
	    		
			
					
	
			 				
	
	
